import collections
import itertools
import ubelt as ub
import warnings
from abc import abstractmethod
from functools import lru_cache
from typing import Iterable, Optional, Dict
from dataclasses import dataclass

try:
    from xdev import profile
except Exception:
    profile = ub.identity


def trackid_is_default(trackid):
    '''
    Hack to decide if a trackid is really a site_id or if it was randomly
    assigned
    '''
    if trackid is None:
        return True
    try:
        int(trackid)
        return True
    except ValueError:
        return False


# Poly = Union[kwimage.Polygon, kwimage.MultiPolygon]


class TrackFunction(collections.abc.Callable):
    '''
    Abstract class that all track functions should inherit from.
    '''

    @abstractmethod
    def __call__(self, sub_dset):
        '''
        Ensure each annotation in coco_dset has a track_id.

        Returns:
            kwcoco.CocoDataset
        '''
        raise NotImplementedError('must be implemented by subclasses')

    def apply_per_video(self, coco_dset, overwrite=False):
        '''
        Main entrypoint for this class.
        '''
        import kwcoco
        legacy = False

        tracked_subdsets = []
        vid_gids = coco_dset.index.vidid_to_gids.values()
        total = len(coco_dset.index.vidid_to_gids)
        for gids in ub.ProgIter(vid_gids,
                                total=total,
                                desc='apply_per_video',
                                verbose=3):

            # Beware, in the past there was a crash here that required
            # wrapping the rest of this loop in a try/except. -csg
            sub_dset = self.safe_apply(coco_dset,
                                       gids,
                                       overwrite,
                                       legacy=legacy)
            if legacy:
                coco_dset = sub_dset
            else:
                tracked_subdsets.append(sub_dset)

        if not legacy:
            # Tracks were either updated or added.
            # In the case they were updated the existing track ids should
            # be disjoint. All new tracks should not overlap with

            _debug = 0

            from watch.utils import kwcoco_extensions
            new_trackids = kwcoco_extensions.TrackidGenerator(None)
            fixed_subdataset = []
            for sub_dset in ub.ProgIter(tracked_subdsets,
                                        desc='Ensure ok tracks',
                                        verbose=3):

                if _debug:
                    sub_dset = sub_dset.copy()

                # Rebuild the index to ensure any hacks are removed.
                # We should be able to remove this step.
                # sub_dset._build_index()

                sub_annots = sub_dset.annots()
                sub_tids = sub_annots.lookup('track_id')
                existing_tids = set(sub_tids)

                collisions = existing_tids & new_trackids.used_trackids
                if _debug:
                    print('existing_tids = {!r}'.format(existing_tids))
                    print('collisions = {!r}'.format(collisions))

                new_trackids.exclude_trackids(existing_tids)
                if collisions:
                    old_tid_to_aids = ub.group_items(sub_annots, sub_tids)
                    assert len(old_tid_to_aids) == len(existing_tids)
                    print(f'Resolve {len(collisions)} track collisions')
                    # Change the track ids of any collisions
                    for old_tid in collisions:
                        new_tid = next(new_trackids)
                        # Note: this does not update the index, but we
                        # are about to clobber it anyway, so it doesnt matter
                        for aid in old_tid_to_aids[old_tid]:
                            ann = sub_dset.index.anns[aid]
                            ann['track_id'] = new_tid
                        existing_tids.add(new_tid)
                new_trackids.exclude_trackids(existing_tids)

                if _debug:
                    after_tids = set(sub_annots.lookup('track_id'))
                    print('collisions = {!r}'.format(collisions))
                    print(f'{after_tids=}')

                fixed_subdataset.append(sub_dset)

            # Is this safe to do? It would be more efficient
            coco_dset = kwcoco.CocoDataset.union(*fixed_subdataset,
                                                 disjoint_tracks=False)

            if _debug:
                x = coco_dset.annots().images.get('video_id')
                y = coco_dset.annots().get('track_id')
                z = ub.group_items(x, y)
                track_to_num_videos = ub.map_vals(set, z)
                if track_to_num_videos:
                    assert max(map(len, track_to_num_videos.values())) == 1, (
                        'track belongs to multiple videos!')
        return coco_dset

    @profile
    def safe_apply(self, coco_dset, gids, overwrite, legacy=True):
        import numpy as np
        DEBUG_JSON_SERIALIZABLE = 0
        if DEBUG_JSON_SERIALIZABLE:
            from watch.utils.util_json import debug_json_unserializable

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(coco_dset.dataset,
                                      'Input to safe_apply: ')

        if legacy:
            sub_dset, rest_dset = self.safe_partition(coco_dset,
                                                      gids,
                                                      remove=True)
        else:
            sub_dset = self.safe_partition(coco_dset, gids, remove=False)

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(sub_dset.dataset, 'Before __call__')
        if overwrite:
            sub_dset = self(sub_dset)
            if DEBUG_JSON_SERIALIZABLE:
                debug_json_unserializable(sub_dset.dataset,
                                          'After __call__ (overwrite)')
        else:
            orig_annots = sub_dset.annots()
            orig_tids = orig_annots.get('track_id', None)
            orig_trackless_flags = np.array([tid is None for tid in orig_tids])
            orig_aids = list(orig_annots)

            # TODO more sophisticated way to check if we can skip self()
            sub_dset = self(sub_dset)
            if DEBUG_JSON_SERIALIZABLE:
                debug_json_unserializable(sub_dset.dataset, 'After __call__')

            # if new annots were not created, rollover the old tracks
            new_annots = sub_dset.annots()
            if new_annots.aids == orig_aids:
                new_tids = new_annots.get('track_id', None)
                # Only overwrite track ids for annots that didn't have them
                new_tids = np.where(orig_trackless_flags, new_tids, orig_tids)

                # Ensure types are json serializable
                import numbers

                def _fixtype(tid):
                    # need to keep strings the same, but integers need to be
                    # case from numpy to python ints.
                    if isinstance(tid, numbers.Integral):
                        return int(tid)
                    else:
                        return tid

                new_tids = list(map(_fixtype, new_tids))

                new_annots.set('track_id', new_tids)

        # TODO: why is this assert here?
        assert None not in sub_dset.annots().lookup('track_id', None)

        if legacy:
            out_dset = self.safe_union(rest_dset, sub_dset)
        else:
            out_dset = sub_dset

        if DEBUG_JSON_SERIALIZABLE:
            debug_json_unserializable(out_dset.dataset,
                                      'Output of safe_apply: ')
        return out_dset

    @staticmethod
    @profile
    def safe_partition(coco_dset, gids, remove=True):
        sub_dset = coco_dset.subset(gids=gids, copy=True)
        # HACK ensure tracks are not duplicated between videos
        # (if they are, this is fixed in dedupe_tracks anyway)
        sub_dset.index.trackid_to_aids.update(coco_dset.index.trackid_to_aids)
        if remove:
            rest_gids = list(set(coco_dset.imgs.keys()) - set(gids))
            rest_dset = coco_dset.subset(rest_gids)
            return sub_dset, rest_dset
        else:
            return sub_dset

    @staticmethod
    @profile
    def safe_union(coco_dset, new_dset, existing_aids=[]):
        coco_dset._build_index()
        new_dset._build_index()
        # we handle tracks in normalize.dedupe_tracks anyway, and
        # disjoint_tracks=True interferes with keeping site_ids around as
        # track_ids.
        # return coco_dset.union(new_dset, disjoint_tracks=True)
        return coco_dset.union(new_dset, disjoint_tracks=False)


class NoOpTrackFunction(TrackFunction):
    '''
    Use existing tracks.
    '''

    def __init__(self, **kwargs):
        self.kwargs = kwargs  # Unused

    def __call__(self, sub_dset):
        return sub_dset


@dataclass
class NewTrackFunction(TrackFunction):
    '''
    Specialization of TrackFunction to create polygons that do not yet exist
    in coco_dset, and add them as new annotations
    '''
    viz_out_dir: Optional[ub.Path] = None

    def __post_init__(self):
        global VIZ_DPATH
        VIZ_DPATH = self.viz_out_dir
        # HACK

    def __call__(self, sub_dset):
        print('Create tracks')
        tracks = self.create_tracks(sub_dset)
        print('Add tracks to dset')
        sub_dset = self.add_tracks_to_dset(sub_dset, tracks)
        print('After tracking sub_dset.stats(): ' +
              ub.urepr(sub_dset.basic_stats()))
        return sub_dset

    @abstractmethod
    def create_tracks(self, sub_dset):
        """
        Args:
            sub_dset (CocoDataset):

        Returns:
            GeoDataFrame
        """
        raise NotImplementedError('must be implemented by subclasses')

    @abstractmethod
    def add_tracks_to_dset(self, sub_dset, tracks):
        """
        Args:
            tracks (GeoDataFrame):

        Returns:
            kwcoco.CocoDataset
        """
        raise NotImplementedError('must be implemented by subclasses')


def check_only_bg(category_sequence, bg_name=['No Activity']):
    if len(set(category_sequence) - set(bg_name)) == 0:
        return True
    else:
        return False


# --- geopandas utils ---
# make this a class?


def gpd_sort_by_gid(gdf, sorted_gids):

    dct = dict(zip(sorted_gids, range(len(sorted_gids))))
    gdf['gid_order'] = gdf['gid'].map(dct)

    # assert gdf['gid'].map(dct).groupby(
    # lambda x: x).is_monotonic_increasing.all()

    if gdf.empty:
        return gdf
    else:
        return gdf.groupby('track_idx', group_keys=False).apply(
            lambda x: x.sort_values('gid_order')).reset_index(drop=True).drop(
                columns=['gid_order'])


def gpd_len(gdf):
    return gdf['track_idx'].nunique()


@profile
def gpd_compute_scores(
        gdf,
        sub_dset,
        thrs: Iterable,
        ks: Dict,
        USE_DASK=False,
        resolution=None):
    import numpy as np
    import pandas as pd

    def compute_scores(grp, thrs=[], keys=[]):
        gid = getattr(grp, 'name', None)
        if gid is None:
            for thr in thrs:
                grp[[(k, thr) for k in keys]] = 0
        else:
            heatmaps = []
            img = sub_dset.coco_image(gid)
            for k in keys:
                # TODO handle keys as channelcodes
                if k in img.channels:
                    heatmap = img.imdelay(k, space='video', resolution=resolution).finalize()
                    heatmap = np.squeeze(heatmap, -1)
                else:
                    w, h = img.imdelay(space='video', resolution=resolution).dsize
                    heatmap = np.zeros((h, w))
                heatmaps.append(heatmap)
            heatmaps = np.stack(heatmaps, axis=0)
            score_cols = list(itertools.product(keys, thrs))
            scores = grp['poly'].apply(
                lambda p: pd.Series(dict(zip(
                    score_cols,
                    # awk, making this serializable for kwcoco dataset
                    list(ub.flatten(score_poly(p, heatmaps, threshold=thrs)))))
                ))
            grp[score_cols] = scores
        return grp

    ks = {k: v for k, v in ks.items() if v}
    _valid_keys = list(set().union(itertools.chain.from_iterable(
        ks.values())))  # | ks.keys()
    score_cols = list(itertools.product(_valid_keys, thrs))

    if USE_DASK:  # 63% runtime
        import dask_geopandas
        # https://github.com/geopandas/dask-geopandas
        # _col_order = gdf.columns  # doesn't matter
        gdf = gdf.set_index('gid')
        # npartitions and chunksize are mutually exclusive
        gdf = dask_geopandas.from_geopandas(gdf, npartitions=8)
        meta = gdf._meta.join(pd.DataFrame(columns=score_cols, dtype=float))
        gdf = gdf.groupby('gid', group_keys=False).apply(compute_scores,
                                                         thrs=thrs,
                                                         keys=_valid_keys,
                                                         meta=meta)
        # raises this, which is probably fine:
        # /home/local/KHQ/matthew.bernstein/.local/conda/envs/watch/lib/python3.9/site-packages/rasterio/features.py:362:
        # NotGeoreferencedWarning: Dataset has no geotransform, gcps, or rpcs.
        # The identity matrix will be returned.
        # _rasterize(valid_shapes, out, transform, all_touched, merge_alg)

        gdf = gdf.compute()  # gdf is now a GeoDataFrame again
        gdf = gdf.reset_index()
        # gdf = gdf.reindex(columns=_col_order)

    else:  # 95% runtime
        grouped = gdf.groupby('gid', group_keys=False)
        gdf = grouped.apply(compute_scores, thrs=thrs, keys=_valid_keys)

    # fill nan scores from nodata pxls
    # groupby track instead of gid
    def _fillna(grp):
        if len(grp) == 0:
            grp = grp.reindex(list(ub.oset(grp.columns) | score_cols), axis=1)
        else:
            grp[score_cols] = grp[score_cols].fillna(method='ffill').fillna(0)
        return grp

    grouped = gdf.groupby('track_idx', group_keys=False)
    scored_gdf = grouped.apply(_fillna)

    # copy over to summed fg/bg channels
    for thr in thrs:
        for k, kk in ks.items():
            if kk:
                # https://github.com/pandas-dev/pandas/issues/20824#issuecomment-384432277
                sum_cols = [(ki, thr) for ki in kk]
                sum_cols = list(ub.oset(scored_gdf.columns) & sum_cols)
                scored_gdf[(k, thr)] = scored_gdf[sum_cols].sum(axis=1)

    return scored_gdf


# -----------------------


@profile
def pop_tracks(coco_dset,
               cnames: Iterable[str],
               remove: bool = True,
               score_chan=None,
               resolution: Optional[str] = None):
    '''
    Convert kwcoco annotations into tracks.

    Args:
        coco_dset (kwcoco.CocoDataset):

        cnames: category names

        remove: remove the annotations from coco_dset

        score_chan (kwcoco.ChannelSpec | None):
            score the track polygons by image overlap with this channel

    Returns:
        gpd dataframe.
        Mutates coco_dset if remove=True.
    '''
    # TODO could refactor to work on coco_dset.annots() and integrate
    import geopandas as gpd
    import numpy as np
    cnames = list(set(cnames))

    annots = coco_dset.annots()
    annots = annots.compress(
        np.in1d(np.array(annots.cnames, dtype=str), cnames))
    if len(annots) < 1:
        print(f'warning: no {cnames} annots in dset {coco_dset.tag}!')

    # Load polygon annotation segmentation in video space
    coco_imgs = annots.images.coco_images
    polys = []
    for coco_img, ann in zip(coco_imgs, annots.objs):
        poly = coco_img._annot_segmentation(ann, space='video')
        polys.append(poly)

    assert len(polys) == len(annots), ('TODO handle multipolygon boundaries')

    polys = [p.to_shapely() for p in polys]
    gdf = gpd.GeoDataFrame(dict(gid=annots.gids,
                                poly=polys,
                                track_idx=annots.get('track_id')),
                           geometry='poly')
    if score_chan is not None:
        keys = {score_chan.spec: list(score_chan.unique())}
        # gdf = gpd_compute_scores(gdf, coco_dset, [-1], keys, USE_DASK=True,
        gdf = gpd_compute_scores(gdf, coco_dset, [-1], keys, USE_DASK=False,
                                 resolution=resolution)
    # TODO standard way to access sorted_gids
    sorted_gids = coco_dset.index._set_sorted_by_frame_index(
        np.unique(annots.gids))
    gdf = gpd_sort_by_gid(gdf, sorted_gids)

    if remove:
        coco_dset.remove_categories(cnames, keep_annots=False)

    return gdf


@lru_cache(maxsize=512)
def _rasterized_poly(shp_poly, h, w, pixels_are):
    import kwimage
    poly = kwimage.MultiPolygon.from_shapely(shp_poly)
    mask = poly.to_mask((h, w), pixels_are=pixels_are).data
    return mask


@profile
def score_poly(poly, probs, threshold=-1, use_rasterio=True):
    '''
    Args:
        poly: kwimage.Polygon or MultiPolygon in pixel coords

        probs: heatmap to compare poly against. Any leading batch dimensions
        will be preserved in output, e.g. (gid chan w h) -> (gid chan)

        use_rasterio: use rasterio.features module instead of kwimage

        threshold: Return fraction of poly with probs > threshold.
        If -1, return average value of probs in poly. Can be a list of values,
        in which case returns all of them.

    '''
    import kwimage
    import numpy as np
    if not isinstance(poly, (kwimage.Polygon, kwimage.MultiPolygon)):
        poly = kwimage.MultiPolygon.from_shapely(poly)  # 2.4% of runtime

    _return_list = isinstance(threshold, Iterable)
    if not _return_list:
        threshold = [threshold]

    # First compute the valid bounds of the polygon
    # And create a mask for only the valid region of the polygon
    box = poly.bounding_box().quantize().to_xywh()
    # Ensure box is inside probs
    ymax, xmax = probs.shape[-2:]
    box = box.clip(0, 0, xmax, ymax).to_xywh()
    if box.area[0][0] == 0:
        warnings.warn(
            'warning: scoring a polygon against an img with no overlap!')
        zeros = np.zeros(probs.shape[:-2])
        return [zeros] * len(threshold) if _return_list else zeros
    x, y, w, h = box.data[0]
    pixels_are = 'areas' if use_rasterio else 'points'
    # kwimage inverse
    # 95% of runtime... would batch be faster?
    rel_poly = poly.translate((-x, -y))
    # rel_mask = rel_poly.to_mask((h, w), pixels_are=pixels_are).data
    # shapely polys hash correctly (based on shape, not memory location)
    # kwimage polys don't
    rel_mask = _rasterized_poly(rel_poly.to_shapely(), h, w, pixels_are)
    # Slice out the corresponding region of probabilities
    rel_probs = probs[..., y:y + h, x:x + w]

    result = []

    # handle nans
    msk = (np.isfinite(rel_probs) * rel_mask).astype(bool)
    for t in threshold:
        if not msk.any():
            result.append(np.nan * np.ones(rel_probs.shape[:-2]))
        elif t == -1:
            mskd = np.ma.array(rel_probs, mask=~msk)
            result.append(mskd.mean(axis=(-2, -1)).filled(0))
        else:
            hard_prob = rel_probs > t
            mskd = np.ma.array(hard_prob, mask=~msk)
            result.append(mskd.mean(axis=(-2, -1)).filled(0))

    return result if _return_list else result[0]


@profile
def mask_to_polygons(probs,
                     thresh,
                     bounds=None,
                     use_rasterio=True,
                     thresh_hysteresis=None):
    """
    Args:
        probs: aka heatmap, image of probability values
        thresh: to turn probs into a hard mask
        bounds: a kwimage or shapely polygon to crop the results to
        use_rasterio: use rasterio.features module instead of kwimage
        thresh_hysteresis: if not None, only keep polygons with at least one
            pixel of score >= thresh_hysteresis

    Returns:
        Iterable[kwcoco.Polygon]

    Example:
        >>> from watch.tasks.tracking.utils import mask_to_polygons
        >>> import kwimage
        >>> probs = kwimage.Heatmap.random(dims=(64, 64),
        >>>                                rng=0).data['class_probs'][0]
        >>> thresh = 0.5
        >>> polys = mask_to_polygons(probs, thresh)
        >>> poly1 = list(polys)[0]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(probs > 0.5)

    Example:
        >>> from watch.tasks.tracking.utils import mask_to_polygons
        >>> import kwimage
        >>> import kwarray
        >>> rng = kwarray.ensure_rng(1043462368)
        >>> probs = kwimage.Heatmap.random(dims=(256, 256), rng=rng,
        >>>                                 ).data['class_probs'][0]
        >>> thresh = 0.5
        >>> polys1 = list(mask_to_polygons(
        >>>             probs, thresh, use_rasterio=0))
        >>> polys2 = list(mask_to_polygons(
        >>>             probs, thresh, use_rasterio=1))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> plt = kwplot.autoplt()
        >>> pnum_ = kwplot.PlotNums(nSubplots=4)
        >>> kwplot.imshow(probs, pnum=pnum_(), title='pixels_are=points')
        >>> for poly in polys1:
        >>>     poly.draw(facecolor='none', edgecolor='kitware_blue', alpha=0.5, linewidth=8)
        >>> kwplot.imshow(probs, pnum=pnum_(), title='pixels_are=areas')
        >>> for poly in polys2:
        >>>     poly.draw(facecolor='none', edgecolor='kitware_green', alpha=0.5, linewidth=8)
    """
    import kwimage
    import numpy as np
    import shapely.geometry
    from scipy.ndimage import label as ndm_label
    # Threshold scores
    if thresh_hysteresis is None:
        binary_mask = (probs > thresh).astype(np.uint8)
    else:
        mask = probs > thresh
        seeds = probs > thresh_hysteresis
        label_img = ndm_label(mask)[0]
        selected = np.unique(np.extract(seeds, label_img))
        binary_mask = np.isin(label_img, selected).astype(np.uint8)

    pixels_are = 'areas' if use_rasterio else 'points'
    if bounds is not None:
        try:  # is this a shapely or geojson object?
            # asShape is being deprecated:
            # https://github.com/shapely/shapely/issues/1100
            bounds = shapely.geometry.shape(bounds)
        except ValueError:  # is this a kwimage object?
            bounds = bounds.to_shapely()
        bounds_mask = kwimage.Polygon.from_shapely(bounds).to_mask(
            probs.shape, pixels_are=pixels_are).numpy().data.astype(np.uint8)
        binary_mask *= bounds_mask

    polygons = kwimage.Mask(
        binary_mask, 'c_mask').to_multi_polygon(pixels_are=pixels_are)

    yield from polygons


def _validate_keys(key, bg_key):
    # for backwards compatibility
    if bg_key is None:
        bg_key = []
    key = list(key) if ub.iterable(key) else [key]
    bg_key = list(bg_key) if ub.iterable(bg_key) else [bg_key]

    # error checking
    if len(key) < 1:
        raise ValueError('must have at least one key')
    if (len(key) > len(set(key)) or len(bg_key) > len(set(bg_key))):
        raise ValueError('keys are duplicated')
    if not set(key).isdisjoint(set(bg_key)):
        raise ValueError('cannot have a key in foreground and background')
    return key, bg_key



"""
CommandLine:
    # Benchmark time sampling
    SMART_DATA_DVC_DPATH=1 XDEV_PROFILE=1 xdoctest -m watch.tasks.fusion.datamodules.spacetime_grid_builder __doc__:0

Example:
    >>> # xdoctest: +REQUIRES(env:SMART_DATA_DVC_DPATH)
    >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
    >>> import watch
    >>> import kwcoco
    >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='ssd')
    >>> coco_fpath = dvc_dpath / 'Drop6/imganns-NZ_R001.kwcoco.zip'
    >>> coco_dset = kwcoco.CocoDataset(coco_fpath)
    >>> window_dims = 128
    >>> time_dims = 5
    >>> builder = SpacetimeGridBuilder(
    >>>     coco_dset,
    >>>     time_dims,
    >>>     window_dims,
    >>>     time_sampling='soft2+distribute',
    >>>     time_kernel='-1y,-8m,-2w,0,2w,8m,1y',
    >>>     keepbound=True,
    >>>     use_annot_info=False,
    >>>     use_grid_positives=False,
    >>>     use_centered_positives=False,
    >>>     respect_valid_regions=False,  # enabling this is slow
    >>>     use_cache=0
    >>> )
    >>> grid = builder.build()
    >>> # xdoctest: +REQUIRES(--show)
    >>> import kwplot
    >>> plt = kwplot.autoplt()
    >>> canvas = visualize_sample_grid(dset, grid, max_vids=1, max_frames=3)
    >>> kwplot.imshow(canvas, doclf=1, fnum=2)
    >>> plt.gca().set_title(ub.codeblock(
        '''
        Sampled using larger scaled windows
        '''))
    >>> kwplot.show_if_requested()
"""

import kwarray
import kwimage
import numpy as np
import ubelt as ub
from watch.utils import util_kwimage
from watch.utils import kwcoco_extensions
from watch import heuristics

try:
    from xdev import profile
except ImportError:
    profile = ub.identity


class SpacetimeGridBuilder:
    """
    A helper class to help build a grid of spacetime windows for a coco
    dataset.

    See :func:`sample_video_spacetime_targets` for the main implementation.
    This will move to a class based approach and ideally be cleaned up as time
    moves on.
    """
    def __init__(
        builder,
        dset,
        time_dims,
        window_dims,
        window_overlap=0.0,
        negative_classes=None,
        keepbound=True,
        include_sensors=None,
        exclude_sensors=None,
        select_images=None,
        select_videos=None,
        time_sampling='hard+distribute',
        time_span=None,
        time_kernel=None,
        use_annot_info=True,
        use_grid_positives=True,
        use_grid_negatives=True,
        use_centered_positives=True,
        window_space_scale=None,
        set_cover_algo=None,
        respect_valid_regions=True,
        workers=0,
        use_cache=1
    ):
        """
        Args:
            dset (kwcoco.CocoDataset): coco dataset

            time_dims (int):
                number of time steps

            window_dims (Tuple[int, int] | str):
                spatial height, width of the sample region or a string code.

            window_overlap (float):
                fractional spatial overlap

            set_cover_algo (str | None):
                Algorithm used to find set cover of image IDs. Options are 'approx' (a greedy solution)
                or 'exact' (an ILP solution). If None is passed, set cover is not computed. The 'exact'
                method requires the packe pulp, available at PyPi.

            window_space_scale (str):
                Code indicating the scale at which to sample. If None uses the
                videospace GSD.

            use_grid_positives (bool):
                if False, will remove any grid sample that contains a positive
                example. In this case use_centered_positives should be True.

            use_grid_negatives (bool | str):
                Use non-annotation locations as negatives. Can be "cleared".

            use_centered_positives (bool):
                extend the grid with extra off-axis samples where positive
                annotations are centered. TODO: we could do a box packing
                to reduce the potential size here.

            use_annot_info (bool):
                if True allows using annotation information to get a better
                train-time grid. Should not be used at test-time.

            time_span (str):
                indicates the desired start/stop date range of the sample

            time_kernel (str):
                mutually exclusive with time span.

            time_sampling (str):
                code for specific temporal sampler: see temporal_sampling.py for
                more information.

            exclude_sensors (List[str]):
                A list of sensors to exclude from the grid

            negative_classes (List[str]):
                indicate class names that should not count towards a region being
                marked as positive.

            respect_valid_regions (bool):
                if True, only place windows in valid regions

            workers (int): parallel workers

            use_cache (bool): uses a disk cache if True
        """
        builder.kw = ub.udict(locals()) - {'builder'}

    def build(builder):
        return sample_video_spacetime_targets(**builder.kw)


@profile
def sample_video_spacetime_targets(dset,
                                   time_dims=None,
                                   window_dims=None,
                                   window_overlap=0.0,
                                   negative_classes=None,
                                   keepbound=True,
                                   include_sensors=None,
                                   exclude_sensors=None,
                                   select_images=None,
                                   select_videos=None,
                                   time_sampling='hard+distribute',
                                   time_span='2y',
                                   time_kernel=None,
                                   use_annot_info=True,
                                   use_grid_positives=True,
                                   use_grid_negatives=True,
                                   use_centered_positives=True,
                                   window_space_scale=None,
                                   set_cover_algo=None,
                                   respect_valid_regions=True,
                                   workers=0,
                                   use_cache=1):
    r"""
    This is the main driver that builds the sample grid.

    The basic idea is that you will slide a spacetime window over the dataset
    and mark where positive andnegative "windows" are. We also put windows
    directly on positive annotations if desired.

    See the above :func:`visualize_sample_grid` for a visualization of what the
    sample grid looks like.

    Ask jon about what the params mean if you need this.
    This code badly needs a refactor.

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> import os
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import watch
        >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='ssd')
        >>> coco_fpath = dvc_dpath / 'Drop6/data_train_split1.kwcoco.zip'
        >>> dset = watch.coerce_kwcoco(coco_fpath)
        >>> window_overlap = 0.0
        >>> window_dims = (128, 128)
        >>> time_dims = 2
        >>> sample_grid = SpacetimeGridBuilder(dset, time_dims, window_dims).build()
        >>> time_sampling = 'hard+distribute'
        >>> positives = list(ub.take(sample_grid['targets'], sample_grid['positives_indexes']))

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> import os
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import watch
        >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='ssd')
        >>> coco_fpath = dvc_dpath / 'Aligned-Drop3-TA1-2022-03-10/combo_LM_nowv_vali.kwcoco.json'
        >>> dset = kwcoco.CocoDataset(coco_fpath)
        >>> window_overlap = 0.5
        >>> time_dims = 2
        >>> window_dims = (128, 128)
        >>> keepbound = False
        >>> exclude_sensors = None
        >>> set_cover_algo = 'approx'
        >>> sample_grid = SpacetimeGridBuilder(dset, time_dims, window_dims, window_overlap, set_cover_algo=set_cover_algo).build()
        >>> time_sampling = 'hard+distribute'
        >>> positives = list(ub.take(sample_grid['targets'], sample_grid['positives_indexes']))

        _ = xdev.profile_now(sample_video_spacetime_targets)(dset, window_dims, window_overlap)

    Example:
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import ndsampler
        >>> import kwcoco
        >>> dset = kwcoco.CocoDataset.demo('vidshapes2-multispectral', num_frames=30)
        >>> window_overlap = 0.0
        >>> time_dims = 3
        >>> window_dims = (256, 256)
        >>> keepbound = False
        >>> time_sampling = 'soft2+distribute'
        >>> sample_grid1 = SpacetimeGridBuilder(
        >>>     dset, time_dims, window_dims, window_overlap,
        >>>     time_sampling='soft2+distribute').build()
        >>> boxes = [kwimage.Boxes.from_slice(target['space_slice'], clip=False).to_xywh() for target in sample_grid1['targets']]
        >>> all_boxes = kwimage.Boxes.concatenate(boxes)
        >>> assert np.all(all_boxes.height == window_dims[0])
        >>> assert np.all(all_boxes.width == window_dims[1])

    Example:
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import ndsampler
        >>> import kwcoco
        >>> dset = kwcoco.CocoDataset.demo('vidshapes2-multispectral', num_frames=30)
        >>> window_overlap = 0.0
        >>> time_dims = 3
        >>> window_dims = (32, 32)
        >>> keepbound = False
        >>> time_sampling = 'soft2+distribute'
        >>> sample_grid1 = SpacetimeGridBuilder(
        >>>     dset, time_dims, window_dims, window_overlap, exclude_sensors='Foo',
        >>>     time_sampling='soft2+distribute').build()
        >>> sample_grid2 = SpacetimeGridBuilder(
        >>>     dset, time_dims, window_dims, window_overlap,
        >>>     time_sampling='contiguous+pairwise').build()

        ub.peek(sample_grid1['vidid_to_time_sampler'].values()).show_summary(fnum=1)
        ub.peek(sample_grid2['vidid_to_time_sampler'].values()).show_summary(fnum=2)
        _ = xdev.profile_now(sample_video_spacetime_targets)(dset, time_dims, window_dims, window_overlap)

        import xdev
        globals().update(xdev.get_func_kwargs(sample_video_spacetime_targets))

    Ignore:
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import watch
        >>> import kwcoco
        >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='ssd')
        >>> coco_fpath = dvc_dpath / 'Drop6/imganns-NZ_R001.kwcoco.zip'
        >>> dset = kwcoco.CocoDataset(coco_fpath)
        >>> window_overlap = 0.0
        bad_image_ids = []
        good_image_ids = []
        for ann in dset.anns.values():
            if 'bbox' not in ann:
                bad_image_ids.append(ann['image_id'])
            else:
                good_image_ids.append(ann['image_id'])
            ...
        dset.videos(ub.oset(dset.images(ub.oset(bad_image_ids)).lookup('video_id'))).lookup('name')
        dset.videos(ub.oset(dset.images(ub.oset(good_image_ids)).lookup('video_id'))).lookup('name')
    """
    if isinstance(window_dims, int):
        window_dims = (window_dims, window_dims)

    if window_dims is not None and isinstance(window_dims, tuple) and len(window_dims) == 3:
        ub.schedule_deprecation(
            'watch', 'window_dims', 'argument in spacetime_grid_builder',
            migration='window_dims no longer supports T,H,W specification, use time_times=T, window_dims=(H, W)',
            deprecate='now')
        winspace_time_dims = window_dims[0]
        winspace_space_dims = window_dims[1:3]
        assert time_dims is None, 'cant do old style and new style'
    else:
        winspace_time_dims = time_dims
        winspace_space_dims = window_dims

    if time_kernel is not None and time_span is not None:
        raise ValueError('time_span and time_kernel are mutually exclusive')

    # from ndsampler import isect_indexer
    # _isect_index = isect_indexer.FrameIntersectionIndex.from_coco(dset)
    parts = set(time_sampling.split('+'))
    affinity_type_parts = parts & {
        'hard', 'hardish', 'contiguous', 'soft2', 'soft', 'hardish2',
        'hardish3', 'soft2-contiguous-hardish3', 'uniform',
        'uniform-soft2-contiguous-hardish3',
        'uniform-soft3-contiguous-hardish3',
        'uniform-soft5-soft4-contiguous',
        'soft4', 'soft5',
    }
    update_rule_parts = parts & {'distribute', 'pairwise'}
    unknown = (parts - affinity_type_parts) - update_rule_parts
    if unknown:
        raise ValueError('Unknown time-sampling parts: {}'.format(unknown))

    affinity_type = '+'.join(list(affinity_type_parts))
    update_rule = '+'.join(list(update_rule_parts))
    if not update_rule:
        update_rule = 'distribute'

    if negative_classes is None:
        negative_classes = heuristics.BACKGROUND_CLASSES

    dset_hashid = dset._cached_hashid()

    # Intersection over smaller area wrt to window vs valid regions.
    refine_iosa_thresh = 0.2  # parametarize?

    # TODO: we can disable respect valid regions here and then just do it on
    # the fly in the dataloader, but it is unclear which is more efficient.

    print(f'winspace_time_dims={winspace_time_dims}')
    print(f'winspace_space_dims={winspace_space_dims}')
    depends = [
        dset_hashid,
        negative_classes,
        affinity_type,
        update_rule,
        winspace_space_dims,
        winspace_time_dims,
        window_overlap,
        window_space_scale,
        negative_classes, keepbound,
        include_sensors,
        exclude_sensors,
        select_videos,
        select_images,
        affinity_type,
        time_span, time_kernel, use_annot_info,
        set_cover_algo,
        use_grid_positives,
        use_grid_negatives,
        use_centered_positives,
        refine_iosa_thresh,
        respect_valid_regions,
        'cache_v13',
    ]
    # Higher level cacher (not sure if adding this secondary level of caching
    # is faster or not).
    dset_name = ub.Path(dset.fpath).name
    cache_dpath = ub.Path.appdir('watch', 'grid_cache').ensuredir()
    cacher = ub.Cacher('sample_grid-dataset-cache_' + dset_name,
                       dpath=cache_dpath, depends=depends, enabled=use_cache,
                       verbose=4)
    sample_grid = cacher.tryload()
    if sample_grid is None:

        # Only build a grid over the selected images / videos
        selected_gids = kwcoco_extensions.filter_image_ids(
            dset, include_sensors=include_sensors,
            exclude_sensors=exclude_sensors,
            select_videos=select_videos,
            select_images=select_images,
        )
        selected_vidid_per_gid = dset.images(selected_gids).lookup('video_id')
        selected_vidid_to_gids = ub.group_items(selected_gids, selected_vidid_per_gid)

        # Given an video
        all_vid_ids = sorted(set(selected_vidid_to_gids.keys()))

        from watch.utils import util_parallel
        workers = util_parallel.coerce_num_workers(workers)
        workers = min(len(all_vid_ids), workers)
        if workers == 1:
            workers = 0
        mode = 'process'
        jobs = ub.JobPool(mode=mode, max_workers=workers)

        # TODO: Reducing the information that needs to be passed to each worker
        # would help improve speed here. The dset itself is the biggest offender.
        verbose = 1 if workers == 0 else 0
        for video_id in ub.ProgIter(all_vid_ids, desc='Submit sample video regions'):
            # Ensure the ordering of image ids is valid
            video_gids = selected_vidid_to_gids[video_id]
            sortx = ub.argsort(dset.images(video_gids).lookup('frame_index'))
            video_gids = list(ub.take(video_gids, sortx))

            job = jobs.submit(
                _sample_single_video_spacetime_targets, dset, dset_hashid,
                video_id, video_gids, winspace_time_dims, winspace_space_dims,
                window_overlap, negative_classes, keepbound,
                affinity_type, update_rule, time_span, time_kernel, use_annot_info,
                use_grid_positives, use_grid_negatives, use_centered_positives, window_space_scale,
                set_cover_algo, use_cache, respect_valid_regions,
                refine_iosa_thresh, verbose)
            job.video_id = video_id

        targets = []
        positive_idxs = []
        negative_idxs = []
        vidid_to_time_sampler = {}
        vidid_to_valid_gids = {}
        vidid_to_meta = ub.ddict(dict)
        for job in jobs.as_completed(desc='Collect region sample grids',
                                     progkw=dict(verbose=3)):
            video_id = job.video_id
            _cached, meta, time_sampler, video_gids = job.result()
            offset = len(targets)
            targets.extend(_cached['video_targets'])
            positive_idxs.extend([idx + offset for idx in _cached['video_positive_idxs']])
            negative_idxs.extend([idx + offset for idx in _cached['video_negative_idxs']])
            vidid_to_time_sampler[video_id] = time_sampler
            vidid_to_valid_gids[video_id] = video_gids
            vidid_to_meta[video_id] = meta

        print('Found {} targets'.format(len(targets)))
        if use_annot_info:
            print('Found {} positives'.format(len(positive_idxs)))
            print('Found {} negatives'.format(len(negative_idxs)))

        sample_grid = {
            'positives_indexes': positive_idxs,
            'negatives_indexes': negative_idxs,
            'targets': targets,
            'vidid_to_valid_gids': vidid_to_valid_gids,
            'vidid_to_time_sampler': vidid_to_time_sampler,
            'vidid_to_meta': vidid_to_meta,
        }
        cacher.save(sample_grid)
    vidid_to_meta = sample_grid['vidid_to_meta']
    from watch.utils.slugify_ext import smart_truncate
    print('vidid_to_meta = {}'.format(smart_truncate(ub.urepr(vidid_to_meta, nl=-1), max_length=1600, head='\n~...', tail='\n...~')))
    return sample_grid


@profile
def _sample_single_video_spacetime_targets(
        dset, dset_hashid, video_id, video_gids, winspace_time_dims, winspace_space_dims,
        window_overlap, negative_classes,
        keepbound, affinity_type, update_rule, time_span, time_kernel,
        use_annot_info, use_grid_positives, use_grid_negatives, use_centered_positives,
        window_space_scale, set_cover_algo, use_cache, respect_valid_regions,
        refine_iosa_thresh, verbose):
    """
    Do this for a single video so we can parallelize.

    Called as the main worker function in
    :func:`sample_video_spacetime_targets`.

    Note this introduces a new temporary space:
        window-space, which is a scaled version of video-space.
        This is only used internally. The final targets are returned in video
        space.
    """
    from watch.tasks.fusion.datamodules import temporal_sampling as tsm  # NOQA
    from watch.tasks.fusion.datamodules import data_utils

    @ub.memoize
    def get_warp_vid_from_img(gid):
        """
        Abstract the transform to bring us into whatever the internal "window
        space" is. Depends on whatever "window_scale" is computed as.
        """
        coco_img = dset.coco_image(gid)
        warp_vid_from_img = coco_img.warp_vid_from_img
        return warp_vid_from_img

    @ub.memoize
    def get_image_valid_region_in_vidspace(gid):
        coco_poly = dset.index.imgs[gid].get('valid_region', None)
        if not coco_poly:
            sh_poly_vid = None
        else:
            warp_vid_from_img = get_warp_vid_from_img(gid)
            kw_poly_img = kwimage.MultiPolygon.coerce(coco_poly)
            kw_poly_vid = kw_poly_img.warp(warp_vid_from_img)
            sh_poly_vid = kw_poly_vid.to_shapely()
        return sh_poly_vid

    # It is important that keepbound is True at test time, otherwise we may not
    # predict on the bottom right of the image.
    keepbound = True
    video_info = dset.index.videos[video_id]
    video_name = video_info['name']

    # Create a box to represent the "window-space" extent, and determine how we
    # are going to slide a window over it.
    vidspace_gsd = video_info.get('target_gsd', None)
    resolved_scale = data_utils.resolve_scale_request(
        request=window_space_scale, data_gsd=vidspace_gsd)
    window_scale = resolved_scale['scale']

    vidspace_time_dims = winspace_time_dims

    time_sampler = tsm.MultiTimeWindowSampler.from_coco_video(
        dset, video_id, gids=video_gids, time_window=vidspace_time_dims,
        affinity_type=affinity_type, update_rule=update_rule, name=video_name,
        time_kernel=time_kernel, time_span=time_span
    )

    gid_arr = np.array(video_gids)
    time_sampler.video_gids = gid_arr
    time_sampler.gid_to_index = ub.udict(enumerate(time_sampler.video_gids)).invert()
    time_sampler.determenistic = True

    # Convert winspace to vidspace and use that for the rest of the function
    vidspace_video_height = video_info['height']
    vidspace_video_width = video_info['width']
    vidspace_full_dims = [vidspace_video_height, vidspace_video_width]
    winspace_full_dims = np.ceil(np.array(vidspace_full_dims) * window_scale)
    if isinstance(winspace_space_dims, str):
        if winspace_space_dims == 'full':
            vidspace_window_dims = vidspace_full_dims
        else:
            raise KeyError(winspace_space_dims)
    else:
        winspace_window_height, winspace_window_width = winspace_space_dims
        winspace_window_box = kwimage.Boxes([
            [0, 0, winspace_window_width, winspace_window_height]], 'xywh')
        # The window is scaled inversely to the data
        # NOTE: This window size can be too big wrt what is actually requested.
        vidspace_window_box = winspace_window_box.scale(1 / window_scale).quantize()
        vidspace_window_height = vidspace_window_box.height.ravel()[0]
        vidspace_window_width = vidspace_window_box.width.ravel()[0]
        vidspace_window_dims = (vidspace_window_height, vidspace_window_width)

    if use_grid_negatives == 'cleared':
        use_grid_negatives = video_info.get('cleared', False)

    depends = [
        dset_hashid,
        negative_classes,
        affinity_type,
        update_rule,
        video_name,
        gid_arr,
        vidspace_window_dims, window_overlap,
        negative_classes, keepbound,
        affinity_type,
        time_span, use_annot_info,
        use_grid_positives,
        use_grid_negatives,
        use_centered_positives,
        refine_iosa_thresh,
        respect_valid_regions,
        set_cover_algo,
        'cache_v11',
    ]

    # Only use the cache if this is probably going to be a slow operation.
    # Otherwise punt, the entire thing gets cached at the end, so the extra
    # disk IO may just slow things down.
    rough_num_windows = np.prod(np.array(vidspace_full_dims) / np.array(vidspace_window_dims))
    rough_num_cells = len(gid_arr) * rough_num_windows
    probably_slow = rough_num_cells > (16 * 30)

    cache_dpath = ub.Path.appdir('watch', 'grid_cache').ensuredir()
    cacher = ub.Cacher('sliding-window-cache-' + video_name,
                       dpath=cache_dpath, depends=depends,
                       enabled=(use_cache and probably_slow))
    _cached = cacher.tryload()
    if _cached is None:

        video_targets = []
        video_positive_idxs = []
        video_negative_idxs = []
        # For each frame, determenistically compute an initial list of which
        # supporting frames we will look at when making a prediction for the
        # "main" frame. Initially this is only based on temporal metadata.  We
        # may modify this later depending on spatial properties.
        main_idx_to_sample_idxs = {
            main_idx: time_sampler.sample(main_idx)
            for main_idx in time_sampler.indexes
        }
        # print('Time index hash: ' + ub.hash_data(time_sampler.indexes))
        # print('Time affinity hash: ' + ub.hash_data(time_sampler.affinity))
        # print('Time time hash: ' + ub.hash_data(time_sampler.unixtimes))
        # print('Time sensor hash: ' + ub.hash_data(time_sampler.sensors))
        # print('First sample hash: ' + ub.hash_data(main_idx_to_sample_idxs))

        main_idx_to_gids = {
            main_idx: list(ub.take(video_gids, sample_idxs))
            for main_idx, sample_idxs in main_idx_to_sample_idxs.items()
        }

        if use_annot_info:
            qtree, tid_to_infos = _build_vidspace_track_qtree(
                dset, video_gids, negative_classes, vidspace_video_width,
                vidspace_video_height, get_warp_vid_from_img)
        else:
            qtree = None
            tid_to_infos = None

        needs_sliding_window = (
            use_grid_negatives or
            use_centered_positives or
            (not use_annot_info)
        )
        if needs_sliding_window:
            # Do a spatial sliding window (in video space) and handle all the
            # temporal stuff for that window in the internal function.
            slider = kwarray.SlidingWindow(vidspace_full_dims,
                                           vidspace_window_dims,
                                           overlap=window_overlap,
                                           keepbound=keepbound,
                                           allow_overshoot=True)
            slices = list(slider)
            num_cells = len(slices) * len(video_gids)
            probably_slow = num_cells > (16 * 30)

            for vidspace_region in ub.ProgIter(slices, desc='Sliding window',
                                               enabled=probably_slow,
                                               verbose=verbose * probably_slow):

                new_targets = _build_targets_in_spatial_region(
                    dset, video_id, vidspace_region, use_annot_info, qtree,
                    main_idx_to_gids, refine_iosa_thresh, time_sampler,
                    get_image_valid_region_in_vidspace, respect_valid_regions,
                    set_cover_algo)
                new_targets = list(new_targets)

                if not use_annot_info:
                    video_targets.extend(new_targets)
                else:
                    for target in new_targets:
                        label = target['label']
                        if label == 'positive_grid':
                            if not use_grid_positives:
                                continue
                            video_positive_idxs.append(len(video_targets))
                        elif label == 'negative_grid':
                            if not use_grid_negatives:
                                continue
                            video_negative_idxs.append(len(video_targets))
                        video_targets.append(target)

        if use_centered_positives and use_annot_info:
            # FIXME: This code is too slow
            # in addition to the sliding window sample, add positive samples
            # centered around each annotation.
            track_infos = list(tid_to_infos.items())
            for tid, infos in ub.ProgIter(track_infos,
                                          desc='Centered annots',
                                          enabled=len(track_infos) > 4 and probably_slow,
                                          verbose=verbose * (len(track_infos) > 4 and probably_slow)):

                new_targets = _build_targets_around_track(
                    video_id, tid, infos, video_gids, vidspace_window_dims,
                    time_sampler)
                new_targets = list(new_targets)
                for target in new_targets:
                    video_positive_idxs.append(len(video_targets))
                    video_targets.append(target)

        _cached = {
            'video_targets': video_targets,
            'video_positive_idxs': video_positive_idxs,
            'video_negative_idxs': video_negative_idxs,
        }
        cacher.save(_cached)

    # Disable determenism in the returned sampler
    time_sampler.determenistic = False
    meta = {
        'video_name': video_name,
        'resolved_scale': resolved_scale,
        'vidspace_window_space_dims': vidspace_window_dims,
        'winspace_window_space_dims': winspace_space_dims,
        'vidspace_full_dims': vidspace_full_dims,
        'winspace_full_dims': winspace_full_dims,
        'num_available_frames': len(time_sampler.indexes),
        'num_samples': len(_cached['video_targets']),
        'num_pos_samples': len(_cached['video_positive_idxs']),
        'num_neg_samples': len(_cached['video_negative_idxs']),
    }
    return _cached, meta, time_sampler, video_gids


@profile
def _build_targets_around_track(video_id, tid, infos, video_gids,
                                vidspace_window_dims, time_sampler):
    """
    Given information about a track, build targets to ensure the network trains
    with it.
    """
    window_height, window_width = vidspace_window_dims
    for info in infos:
        main_gid = info['gid']
        vidspace_ann_box = kwimage.Boxes([info['vidspace_box']], 'tlbr')
        vidspace_ann_box = vidspace_ann_box.quantize()
        vidspace_ann_box = vidspace_ann_box.resize(width=window_width, height=window_height)
        #  FIXME, this code is ugly
        # TODO: we could make frames where the phase transitions
        # more likely here.
        _hack_main_idx = np.where(time_sampler.video_gids == main_gid)[0][0]
        _sample_idxs = time_sampler.sample(_hack_main_idx)
        sample_gids = list(ub.take(video_gids, _sample_idxs))
        _hack = {_hack_main_idx: sample_gids}
        _hack2 = _hack
        if _hack2:
            gids = _hack2[_hack_main_idx]
            label = 'positive_center'
            vidspace_region = vidspace_ann_box.to_slices()[0]

            target = {
                'main_idx': _hack_main_idx,
                'video_id': video_id,
                'gids': gids,
                'main_gid': main_gid,
                'space_slice': vidspace_region,
                'label': label,
                'resampled': -1,
            }
            yield target


@profile
def _build_targets_in_spatial_region(dset, video_id, vidspace_region,
                                     use_annot_info, qtree, main_idx_to_gids,
                                     refine_iosa_thresh, time_sampler,
                                     get_image_valid_region_in_vidspace,
                                     respect_valid_regions, set_cover_algo):
    """
    Called for each spatial grid in the sliding window. This adds multiple
    targets

    Called as part of :func:`_sample_single_video_spacetime_targets`.
    """
    from watch.tasks.fusion.datamodules import temporal_sampling as tsm  # NOQA
    y_sl, x_sl = vidspace_region
    vidspace_box = kwimage.Boxes.from_slice(vidspace_region).to_ltrb()

    # Find all annotations that pass through this spatial region
    if use_annot_info:
        query = vidspace_box.data[0]
        isect_aids = list(qtree.intersect(query))
        isect_gids = set(dset.annots(isect_aids).lookup('image_id'))

    if respect_valid_regions:
        # Reselect the keyframes if we overlap an invalid region (as
        # denoted in the metadata, further filtering may happen later)
        # todo: refactor to be cleaner
        try:
            main_idx_to_gids2, resampled = _refine_time_sample(
                dset, main_idx_to_gids, vidspace_box,
                refine_iosa_thresh, time_sampler,
                get_image_valid_region_in_vidspace)
        except tsm.TimeSampleError as ex:
            # Hack, just skip the region
            # We might be able to sample less and still be ok
            print(f'TSM ex={ex}')
            raise
            # continue
    else:
        main_idx_to_gids2 = main_idx_to_gids
        resampled = False

    if set_cover_algo is not None:
        # TODO: allow there to be a slightly denser overlapped sampling
        # debug = True
        debug = 0
        if debug:
            print('before applying set cover, len of main_idx_to_gids2', len(main_idx_to_gids2))
        main_idx_to_gids2 = kwarray.setcover(main_idx_to_gids2, algo=set_cover_algo)
        if debug:
            print('after applying set cover', len(main_idx_to_gids2))

    for main_idx, gids in main_idx_to_gids2.items():
        main_gid = time_sampler.video_gids[main_idx]
        label = 'unknown'

        if use_annot_info:
            if isect_aids:
                has_annot = bool(isect_gids & set(gids))
            else:
                has_annot = False
            if has_annot:
                label = 'positive_grid'
            else:
                # Hack: exclude all annotated regions from negative sampling
                label = 'negative_grid'

        target = {
            'main_idx': main_idx,
            'video_id': video_id,
            'gids': gids,
            'main_gid': main_gid,
            'space_slice': vidspace_region,
            'resampled': resampled,
            'label': label,
        }
        yield target


@profile
def _build_vidspace_track_qtree(dset, video_gids, negative_classes,
                                vidspace_video_width, vidspace_video_height,
                                get_warp_vid_from_img):
    """
    Build a data structure that allows for fast lookup of which annotations
    exist in the in the requested "Window Space".

    Called as part of :func:`_sample_single_video_spacetime_targets`.
    """
    import pyqtree
    qtree = pyqtree.Index((0, 0, vidspace_video_width, vidspace_video_height))
    tid_to_infos = ub.ddict(list)
    video_images = dset.images(video_gids)
    video_coco_images = video_images.coco_images
    video_aids = video_images.annots.lookup('id')
    for aids, coco_img in zip(video_aids, video_coco_images):
        img_info = coco_img.img
        gid = img_info['id']
        frame_index = img_info['frame_index']
        annots = dset.annots(aids)
        tids = annots.lookup('track_id', None)
        cids = annots.lookup('category_id', None)
        cnames = dset.categories(cids).name

        warp_vid_from_img = get_warp_vid_from_img(gid)

        # Gather data in imagespace so we can group expensive steps together
        imgspace_xywh = []
        infos = []
        for tid, aid, cid, cname in zip(tids, aids, cids, cnames):
            if cname not in negative_classes:
                imgspace_xywh.append(dset.index.anns[aid]['bbox'])
                infos.append({
                    'gid': gid,
                    'cid': cid,
                    'tid': tid,
                    'frame_index': frame_index,
                    'cname': dset._resolve_to_cat(cid)['name'],
                    'aid': aid,
                })

        imgspace_boxes = kwimage.Boxes(np.array(imgspace_xywh), 'xywh')
        vidspace_boxes = imgspace_boxes.warp(warp_vid_from_img)
        vidspace_boxes = vidspace_boxes.clip(
            0, 0, vidspace_video_width, vidspace_video_height)
        isvalid_flags = vidspace_boxes.area.ravel() > 0
        valid_vidspace_boxes = vidspace_boxes.compress(isvalid_flags)
        valid_infos = list(ub.compress(infos, isvalid_flags))

        for info, ltrb in zip(valid_infos, valid_vidspace_boxes.to_ltrb().data):
            tid = info['tid']
            aid = info['aid']
            if tid is not None:
                info.update({
                    'vidspace_box': ltrb,
                })
                tid_to_infos[tid].append(info)
            # A faster qtree insert would be helpful.
            # This takes 50% of the time in spacetime grid building.
            qtree.insert(aid, ltrb)

    return qtree, tid_to_infos


@profile
def _refine_time_sample(dset, main_idx_to_gids, vidspace_box, refine_iosa_thresh, time_sampler, get_image_valid_region_in_vidspace):
    """
    Refine the time sample based on spatial information
    """
    from watch.tasks.fusion.datamodules import temporal_sampling as tsm  # NOQA
    video_gids = time_sampler.video_gids

    gid_to_isbad = {}
    for gid in video_gids:
        vidspace_valid_poly = get_image_valid_region_in_vidspace(gid)
        gid_to_isbad[gid] = False
        # If the area is of the valid polygon is less than zero, there was
        # probably an issue. treat it as if it didn't specify a valid
        # region.
        if vidspace_valid_poly is not None and vidspace_valid_poly.area > 0:
            vidspace_box_poly = vidspace_box.to_shapely()[0]
            # Intersection over smaller area
            isect = vidspace_valid_poly.intersection(vidspace_box_poly)
            iosa = isect.area / min(vidspace_box_poly.area, vidspace_valid_poly.area)
            if iosa < refine_iosa_thresh:
                gid_to_isbad[gid] = True

    all_bad_gids = [gid for gid, flag in gid_to_isbad.items() if flag]

    rng = kwarray.ensure_rng(1093881655714)

    try:
        resampled = 0
        refined_sample = {}
        for main_idx, gids in main_idx_to_gids.items():
            main_gid = video_gids[main_idx]
            # Skip the sample when the "main" frame is bad.
            if not gid_to_isbad[main_gid]:
                good_gids = [gid for gid in gids if not gid_to_isbad[gid]]
                if good_gids != gids:
                    include_idxs = np.where(kwarray.isect_flags(video_gids, good_gids))[0]
                    exclude_idxs = np.where(kwarray.isect_flags(video_gids, all_bad_gids))[0]
                    chosen = time_sampler.sample(
                        include=include_idxs, exclude=exclude_idxs,
                        error_level=1, return_info=False, rng=rng)
                    new_gids = list(ub.take(video_gids, chosen))
                    # Are we allowed to return less than the initial expected
                    # number of frames? For transformers yes, but we should be
                    # careful to ask the user if they expect this.
                    new_are_bad = [g for g in new_gids if gid_to_isbad[g]]
                    if not new_are_bad:
                        resampled += 1
                        refined_sample[main_idx] = new_gids
                else:
                    refined_sample[main_idx] = gids
    except tsm.TimeSampleError:
        print('had a hard time resampling')
        raise

    # print(f'number of resampled grid cells={resampled}')
    return refined_sample, resampled


def lookup_track_info(coco_dset, tid):
    """
    UNUSED. DEPRECATED OR FIND USE.

    Find the spatio-temporal extent of a track
    """
    track_aids = coco_dset.index.trackid_to_aids[tid]
    vidspace_boxes = []
    track_gids = []
    for aid in track_aids:
        ann = coco_dset.index.anns[aid]
        gid = ann['image_id']
        track_gids.append(gid)
        img = coco_dset.index.imgs[gid]
        bbox = ann['bbox']
        vid_from_img = kwimage.Affine.coerce(img.get('warp_img_to_vid', None))
        imgspace_box = kwimage.Boxes([bbox], 'xywh')
        vidspace_box = imgspace_box.warp(vid_from_img)
        vidspace_boxes.append(vidspace_box)
    all_vidspace_boxes = kwimage.Boxes.concatenate(vidspace_boxes)
    full_vid_box = all_vidspace_boxes.bounding_box().to_xywh()

    frame_index = coco_dset.images(track_gids).lookup('frame_index')
    track_gids = list(ub.take(track_gids, ub.argsort(frame_index)))

    track_info = {
        'tid': tid,
        'full_vid_box': full_vid_box,
        'track_gids': track_gids,
    }
    return track_info


def make_track_based_spatial_samples(coco_dset):
    """
    UNUSED. DEPRECATED OR FIND USE.

    Ignore:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> import os
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import watch
        >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='ssd')
        >>> coco_fpath = dvc_dpath / 'Drop1-Aligned-L1/data.kwcoco.json'
        >>> coco_dset = kwcoco.CocoDataset(coco_fpath)
    """
    tid_list = list(coco_dset.index.trackid_to_aids.keys())
    tid_to_trackinfo = {}
    for tid in tid_list:
        track_info = lookup_track_info(coco_dset, tid)
        gid = track_info['track_gids'][0]
        vidid = coco_dset.index.imgs[gid]['video_id']
        track_info['vidid'] = vidid
        tid_to_trackinfo[tid] = track_info

    vidid_to_tracks = ub.group_items(tid_to_trackinfo.values(), key=lambda x: x['vidid'])

    winspace_space_dims = [96, 96]

    for vidid, trackinfos in vidid_to_tracks.items():
        positive_boxes = []
        for track_info in trackinfos:
            boxes = track_info['full_vid_box']
            positive_boxes.append(boxes.to_cxywh())
        positives = kwimage.Boxes.concatenate(positive_boxes)
        positives_samples = positives.to_cxywh()
        positives_samples.data[:, 2] = winspace_space_dims[0]
        positives_samples.data[:, 3] = winspace_space_dims[1]
        print('positive_boxes = {}'.format(ub.urepr(positive_boxes, nl=1)))

        video = coco_dset.index.videos[vidid]
        full_dims = [video['height'], video['width']]
        window_overlap = 0.0
        keepbound = 0

        window_dims_ = full_dims if winspace_space_dims == 'full' else winspace_space_dims
        slider = kwarray.SlidingWindow(full_dims, window_dims_,
                                       overlap=window_overlap,
                                       keepbound=keepbound,
                                       allow_overshoot=True)

        sliding_boxes = kwimage.Boxes.concatenate(list(map(kwimage.Boxes.from_slice, slider)))
        ious = sliding_boxes.ious(positives)
        overlaps = ious.sum(axis=1)
        negative_boxes = sliding_boxes.compress(overlaps == 0)

        if 1:
            import kwplot
            kwplot.autompl()
            fig = kwplot.figure(fnum=vidid)
            ax = fig.gca()
            ax.set_title(video['name'])
            negative_boxes.draw(setlim=1, color='red', fill=True)
            positives.draw(color='limegreen')
            positives_samples.draw(color='green')


def visualize_sample_grid(dset, sample_grid, max_vids=2, max_frames=6):
    r"""
    Debug visualization for sampling grid

    Draws multiple frames.

    Places a red dot where there is a negative sample (at the center of the negative window)

    Places a blue dot where there is a positive sample

    Draws a yellow polygon over invalid spatial regions.

    Notes:
        * Dots are more intense when there are more temporal coverage of that dot.

        * Dots are placed on the center of the window.  They do not indicate its extent.

        * Dots are blue if they overlap any annotation in their temporal region
          so they may visually be near an annotation.

    Example:
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> from watch.demo.smart_kwcoco_demodata import demo_kwcoco_multisensor
        >>> dset = coco_dset = demo_kwcoco_multisensor(num_frames=3, dates=True, geodata=True, heatmap=True, rng=10)
        >>> window_overlap = 0.0
        >>> window_dims = (32, 32)
        >>> keepbound = False
        >>> time_sampling = 'soft2+distribute'
        >>> use_centered_positives = True
        >>> use_grid_positives = 1
        >>> time_dims = 3
        >>> sample_grid = sample_video_spacetime_targets(
        >>>     dset, time_dims, window_dims, window_overlap, time_sampling=time_sampling,
        >>>     use_grid_positives=use_grid_positives, use_centered_positives=use_centered_positives)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> canvas = visualize_sample_grid(dset, sample_grid, max_vids=1,
        >>>                                max_frames=3)
        >>> kwplot.imshow(canvas, doclf=1)
        >>> plt.gca().set_title(ub.codeblock(
            '''
            Places a red dot where there is a negative sample (at the center of the negative window)

            Places a blue dot where there is a positive sample

            Draws a yellow polygon over invalid spatial regions.
            '''))
        >>> kwplot.show_if_requested()
        >>> #
        >>> # Now demo this same grid, but where we are sampling at a different resolution
        >>> window_space_scale = 0.3
        >>> sample_grid2 = sample_video_spacetime_targets(
        >>>     dset, window_dims, window_overlap, time_sampling=time_sampling,
        >>>     use_grid_positives=use_grid_positives,
        >>>     use_centered_positives=use_centered_positives,
        >>>     window_space_scale=window_space_scale)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> canvas = visualize_sample_grid(dset, sample_grid2, max_vids=1,
        >>>                                max_frames=3)
        >>> kwplot.imshow(canvas, doclf=1, fnum=2)
        >>> plt.gca().set_title(ub.codeblock(
            '''
            Sampled using larger scaled windows
            '''))
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from watch.tasks.fusion.datamodules.spacetime_grid_builder import *  # NOQA
        >>> import watch
        >>> # dset = coco_dset = demo_kwcoco_multisensor(dates=True, geodata=True, heatmap=True)
        >>> dvc_dpath = watch.find_dvc_dpath(hardware='ssd', tags='phase2_data')
        >>> #coco_fpath = dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/combo_DILM_train.kwcoco.json'
        # >>> coco_fpath = dvc_dpath / 'Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/data_vali.kwcoco.json'
        >>> coco_fpath = dvc_dpath / 'Drop6/data_vali_split1.kwcoco.zip'
        >>> big_dset = kwcoco.CocoDataset(coco_fpath)
        >>> vid_gids = big_dset.videos(names=['KR_R002']).images.lookup('id')[0]
        >>> idxs = np.linspace(0, len(vid_gids) - 1, 12).round().astype(int)
        >>> vid_gids = list(ub.take(vid_gids, idxs))
        >>> dset = big_dset.subset(vid_gids)
        >>> window_overlap = 0.0
        >>> window_dims = (3, 256, 256)
        >>> keepbound = False
        >>> time_sampling = 'soft2+distribute'
        >>> use_centered_positives = 0
        >>> use_grid_positives = 1
        >>> window_space_scale = '30GSD'
        >>> sample_grid = sample_video_spacetime_targets(
        >>>     dset, window_dims, window_overlap, time_sampling=time_sampling,
        >>>     use_grid_positives=True,
        >>>     use_centered_positives=use_centered_positives,
        >>>     window_space_scale=window_space_scale)
        >>> print(list(ub.unique([t['space_slice'] for t in sample_grid['targets']], key=ub.hash_data)))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> canvas = visualize_sample_grid(dset, sample_grid, max_vids=1, max_frames=12)
        >>> kwplot.imshow(canvas, doclf=1)
        >>> kwplot.show_if_requested()
        >>> plt.gca().set_title(ub.codeblock(
        >>>     f'''
        >>>     Sample window {window_dims} @ {window_space_scale}
        >>>     '''))
    """
    # Visualize the sample grid
    import pandas as pd
    targets = pd.DataFrame(sample_grid['targets'])

    dataset_canvases = []

    # max_vids = 2
    # max_frames = 6

    vidid_to_videodf = dict(list(targets.groupby('video_id')))

    orientation = 1

    for vidid, video_df in vidid_to_videodf.items():
        video = dset.index.videos[vidid]
        vidname = video['name']
        gid_to_infos = ub.ddict(list)
        for _, row in video_df.iterrows():
            for gid in row['gids']:
                gid_to_infos[gid].append({
                    'gid': gid,
                    'space_slice': row['space_slice'],
                    'label': row['label'],
                })

        video_canvases = []
        common = ub.oset(dset.images(vidid=vidid)) & (gid_to_infos)

        if True:
            # HACK: Use a temporal sampler once to get a nice overview of the
            # dataset in time.
            from dateutil import parser
            from watch.tasks.fusion.datamodules import temporal_sampling as tsm  # NOQA
            images = dset.images(common)
            datetimes = [None if date is None else parser.parse(date) for date in images.lookup('date_captured', None)]
            unixtimes = np.array([np.nan if dt is None else dt.timestamp() for dt in datetimes])
            sensors = images.lookup('sensor_coarse', None)
            time_sampler = tsm.MultiTimeWindowSampler(
                unixtimes=unixtimes, sensors=sensors, time_window=max_frames,
                time_span='1y', affinity_type='hardish3',
                update_rule='distribute+pairwise')
            sample = time_sampler.sample()
            common = list(ub.take(common, sample))

        for gid in common:
            infos = gid_to_infos[gid]
            label_to_items = ub.group_items(infos, key=lambda x: x['label'])
            video = dset.index.videos[vidid]

            shape = (video['height'], video['width'], 4)
            canvas = np.zeros(shape, dtype=np.float32)
            shape = (2, video['height'], video['width'])
            accum = np.zeros(shape, dtype=np.float32)

            for label, items in label_to_items.items():
                label_idx = {'positive_grid': 1, 'positive_center': 1,
                             'negative_grid': 0}[label]
                for info in items:
                    space_slice = info['space_slice']
                    y_sl, x_sl = space_slice
                    # ww = x_sl.stop - x_sl.start
                    # wh = y_sl.stop - y_sl.start
                    ss = accum[(label_idx,) + space_slice].shape
                    if np.prod(ss) > 0:
                        vals = util_kwimage.upweight_center_mask(ss)
                        vals = np.maximum(vals, 0.1)
                        accum[(label_idx,) + space_slice] += vals
                        # Add extra weight to borders for viz
                        accum[label_idx, y_sl.start:y_sl.start + 1, x_sl.start: x_sl.stop] += 0.15
                        accum[label_idx, y_sl.stop - 1:y_sl.stop, x_sl.start:x_sl.stop] += 0.15
                        accum[label_idx, y_sl.start:y_sl.stop, x_sl.start: x_sl.start + 1] += 0.15
                        accum[label_idx, y_sl.start:y_sl.stop, x_sl.stop - 1: x_sl.stop] += 0.15

            neg_accum = accum[0]
            pos_accum = accum[1]

            neg_alpha = neg_accum / (neg_accum.max() * 2)
            pos_alpha = pos_accum / (pos_accum.max() * 2)
            bg_canvas = canvas.copy()
            bg_canvas[..., 0:4] = [0., 0., 0., 1.0]
            pos_canvas = canvas.copy()
            neg_canvas = canvas.copy()
            pos_canvas[..., 0:3] = kwimage.Color('dodgerblue').as01()
            neg_canvas[..., 0:3] = kwimage.Color('orangered').as01()
            neg_canvas[..., 3] = neg_alpha
            pos_canvas[..., 3] = pos_alpha
            neg_canvas = np.nan_to_num(neg_canvas)
            pos_canvas = np.nan_to_num(pos_canvas)

            warp_vid_from_img = kwimage.Affine.coerce(
                dset.index.imgs[gid]['warp_img_to_vid'])

            vid_poly = kwimage.Boxes([[0, 0, video['width'], video['height']]], 'xywh').to_polygons()[0]
            coco_poly = dset.index.imgs[gid].get('valid_region', None)
            if coco_poly is None:
                kw_invalid_poly = None
            else:
                kw_poly_img = kwimage.MultiPolygon.coerce(coco_poly)
                valid_poly = kw_poly_img.warp(warp_vid_from_img)
                sh_invalid_poly = vid_poly.to_shapely().difference(valid_poly.to_shapely())
                kw_invalid_poly = kwimage.MultiPolygon.coerce(sh_invalid_poly)

            final_canvas = kwimage.overlay_alpha_layers([pos_canvas, neg_canvas, bg_canvas])
            final_canvas = kwimage.ensure_uint255(final_canvas)

            annot_dets = dset.annots(gid=gid).detections
            vid_annot_dets = annot_dets.warp(warp_vid_from_img)

            if 1:
                final_canvas = vid_annot_dets.draw_on(
                    final_canvas, color='white', alpha=0.5, kpts=False,
                    labels=False)

            if kw_invalid_poly is not None:
                final_canvas = kw_invalid_poly.draw_on(final_canvas, color='yellow', alpha=0.5)

            # from watch import heuristics
            img = dset.index.imgs[gid]
            header_lines = heuristics.build_image_header_text(
                img=img, vidname=vidname)
            header_text = '\n'.join(header_lines)

            final_canvas = kwimage.draw_header_text(final_canvas, header_text, fit=True)
            video_canvases.append(final_canvas)

            if len(video_canvases) >= max_frames:
                break

        if max_vids == 1:
            video_canvas = kwimage.stack_images_grid(
                video_canvases, axis=orientation, pad=10)
        else:
            video_canvas = kwimage.stack_images(video_canvases, axis=1 - orientation, pad=10)
        dataset_canvases.append(video_canvas)
        if len(dataset_canvases) >= max_vids:
            break

    dataset_canvas = kwimage.stack_images(dataset_canvases, axis=orientation, pad=20)
    if 0:
        import kwplot
        kwplot.autompl()
        kwplot.imshow(dataset_canvas, doclf=1)
    return dataset_canvas



#!/usr/bin/env python3
import geojson
import json
import os
import ubelt as ub
import scriptconfig as scfg


class ScoreTracksConfig(scfg.DataConfig):
    """
    Score and Convert KWCOCO to IARPA GeoJSON
    """
    in_file = scfg.Value(None, required=True, help='Input KWCOCO to convert',
                         position=1)

    images_kwcoco = scfg.Value(None, required=True, help=ub.paragraph(
        '''
            kwcoco file with images where to score polygons.
            '''), group='track scoring')

    threshold = scfg.Value(0.4, help=ub.paragraph(
        '''
            threshold to filter polygons, very sensitive
            '''), group='track scoring')

    model_fpath = scfg.Value(None, help='Path to the depthPCD site validation model')

    region_id = scfg.Value(None, help=ub.paragraph(
        '''
            ID for region that sites belong to. If None, try to infer
            from kwcoco file.
            '''), group='convenience')

    out_kwcoco = scfg.Value(None, help=ub.paragraph(
        '''
            The file path to write the "tracked" kwcoco file to.
            '''))

    out_sites_dir = scfg.Value(None, help=ub.paragraph(
        '''
        The directory where site model geojson files will be written.
        '''))

    out_site_summaries_dir = scfg.Value(None, help=ub.paragraph(
        '''
        The directory path where site summary geojson files will be written.
        '''))

    out_sites_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The file path where a manifest of all site models will be written.
        '''))

    out_site_summaries_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The file path where a manifest of all site summary geojson files will
        be written.
        '''))

    append_mode = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Append sites to existing region GeoJSON.
        '''), group='behavior')


def score_tracks(coco_dset, imCoco_dset, thresh, model_fpath):
    from watch.tasks.depthPCD.model import getModel, normalize, TPL_DPATH

    import numpy as np
    import cv2
    import kwimage
    import ndsampler
    import pandas as pd
    from tqdm import tqdm

    print('loading site validation model')
    proto_fpath = TPL_DPATH / 'deeplab2/max_deeplab_s_backbone_os16.textproto'
    model = getModel(proto=proto_fpath)

    model.load_weights(model_fpath, by_name=True, skip_mismatch=True)
    # model.load_weights('/media/hdd2/antonio/models/urbanTCDs-use.h5', by_name=True, skip_mismatch=True)

    to_keep = []
    for coco_img in imCoco_dset.images().coco_images:
        if coco_img['sensor_coarse'] == 'WV' and 'red' in coco_img.channels:
            to_keep.append(coco_img['id'])

    dset = imCoco_dset.subset(to_keep)

    sampler = ndsampler.CocoSampler(dset)

    # FIXME: pandas is slow, we likely want to use an alternative impl
    imgs = pd.DataFrame(coco_dset.dataset["images"])
    if "timestamp" not in imgs.columns:
        imgs["timestamp"] = imgs["id"]

    annots = pd.DataFrame(coco_dset.dataset["annotations"])

    if annots.shape[0] == 0:
        print("Nothing to filter")
        return coco_dset

    annots = annots[[
        "id", "image_id", "track_id", "score"
    ]].join(
        imgs[["timestamp"]],
        on="image_id",
    )

    track_ids_to_drop = []
    ann_ids_to_drop = []

    regions = []
    tannots = annots.groupby('track_id', axis=0)
    tq = tqdm(total=len(tannots))
    for track_id, track_group in tannots:

        first_annot_id = track_group["id"].tolist()[0]
        first_image_id = track_group["image_id"].tolist()[0]
        first_coco_img = coco_dset.coco_image(first_image_id)
        first_annot = coco_dset.anns[first_annot_id]
        imgspace_annot_box = kwimage.Box.coerce(first_annot['bbox'], format='xywh')
        vidspace_annot_box = imgspace_annot_box.warp(first_coco_img.warp_vid_from_img)
        ref_coco_img = first_coco_img  # dset.coco_image(dset.dataset['images'][0]['id'])

        if 1:
            # Because we want a higher resolution, we need to scale the requested
            # videospace region down. Looks like quantization errors may happen
            # here not sure how I deal with in the dataloader, it probably needs to
            # be fixed there too.

            res = '2GSD'
            scale_res_from_vidspace = ref_coco_img._scalefactor_for_resolution(space='video', resolution=res)

            # cxy = vidspace_annot_box.to_cxywh().data[0:2]
            # warp_res_from_vidspace = kwimage.Affine.coerce(scale=scale_res_from_vidspace, about=cxy)
            warp_res_from_vidspace = kwimage.Affine.scale(scale_res_from_vidspace)

            # Convert the video space annotation into requested resolution "window space"
            winspace_annot_box = vidspace_annot_box.warp(warp_res_from_vidspace)

            # Convert to center xy/with/height format
            winspace_annot_box = winspace_annot_box.toformat('cxywh')

            # Force the box to be a specific size at our window resolution
            # THIS IS THE BUG
            # winspace_target_box = winspace_annot_box.resize(*force_dsize)

            # Workaround
            winspace_target_box = winspace_annot_box.copy()
            #            size = max(winspace_target_box.data[2], winspace_target_box.data[3])
            winspace_target_box.data[2:4] = (224, 224)

            # Convert the box back to videospace
            vidspace_target_box = winspace_target_box.warp(warp_res_from_vidspace.inv())

            # Get the slice for video space
            vidspace_slice = vidspace_target_box.quantize().to_slice()
        else:
            vidspace_slice = vidspace_annot_box.to_slice()

        # find in second video
        videoName = first_coco_img.video['name']
        vidid = -1
        for v in sampler.dset.index.videos.values():
            if v['name'] == videoName:
                vidid = v['id']
        if vidid == -1:
            tq.update(1)
            continue
        target = {
            'vidid': vidid,
            'channels': 'blue|green|red',
            'allow_augment': False,
            'space_slice': vidspace_slice,
            'use_native_scale': True,
        }
        data = sampler.load_sample(target, with_annots=False, visible_thres=1.0)

        ims = data['im']

        good_ims = []
        for i in ims:
            im = np.stack([ii[..., 0] for ii in i], axis=-1) / (4096 * 2)
            im[im < 0] = 0
            im[im > 1] = 1
            im = (im * 255).astype(np.uint8)
            if im.shape[-1] == 3:
                im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
            if np.mean(im == 0) > .2:
                continue
            good_ims.append(normalize(im))

        # a little average at start vs end
        nAvg = 2
        if len(good_ims) < nAvg + 1:
            tq.update(1)
            continue
        ims = []
        for i in range(nAvg):
            for j in range(-1, -nAvg - 1, -1):
                first = good_ims[i]
                last = good_ims[j]
                ims.append(np.stack([first, 0.5 * first + 0.5 * last, last], axis=-1).astype(np.float32))

        score = np.mean(model.predict(np.array(ims), batch_size=1, verbose=False)[8])
        tq.set_description('%s-%d score %3.2f' % (videoName, track_id, score))
        tq.update(1)
        if 0:

            cv2.namedWindow('main', cv2.WINDOW_GUI_NORMAL)
            i = 1
            first = good_ims[0]
            last = good_ims[-1]
            while 1:
                if i == 1:
                    i = 0
                    im = first
                else:
                    i = 1
                    im = last
                cv2.imshow('main', im / 255)
                q = cv2.waitKey(0)
                if q == 113:
                    break

        #        ks = list(coco_dset.index.videos.keys())

        if score < thresh:  # or coco_dset.index.videos[ks[0]]['name'] == 'AE_R001':
            track_ids_to_drop.append(track_id)
            ann_ids_to_drop.extend(track_group["id"].tolist())
            regions.append(videoName)

    print(f"Dropping {len(ann_ids_to_drop)} annotations from {len(track_ids_to_drop)} tracks.")
    from collections import Counter
    print(Counter(regions))
    if len(ann_ids_to_drop) > 0:
        coco_dset.remove_annotations(ann_ids_to_drop)

    return coco_dset


def main(**kwargs):
    args = ScoreTracksConfig.cli(cmdline=True, data=kwargs, strict=True)
    import rich
    rich.print('args = {}'.format(ub.urepr(args, nl=1)))

    # Import this first
    print('Importing tensorflow stuff (can take a while)')
    from watch.tasks.depthPCD.model import getModel, normalize, TPL_DPATH  # NOQA

    from watch.cli.kwcoco_to_geojson import convert_kwcoco_to_iarpa, create_region_header
    from watch.utils import process_context

    import kwcoco
    from kwcoco.util import util_json

    if args.model_fpath is None:
        print('warning: the path to the model was not explicitly specified, '
              'attempting to automatically infer it')
        cand_model_path = ub.Path(os.environ.get('DVC_EXPT_DPATH', '')) / 'models/depthPCD/basicModel2.h5'
        if cand_model_path.exists():
            args.model_fpath = cand_model_path
        else:
            raise IOError(
                f'Attempted to infer model path {cand_model_path}, '
                'but it does not exist. Please specify it explicitly')

    model_fpath = ub.Path(args.model_fpath)
    if not model_fpath.exists():
        raise IOError(f'Specified {model_fpath=} does not exist')

    coco_dset = kwcoco.CocoDataset.coerce(args.in_file)

    tracking_output = {
        'type': 'tracking_result',
        'info': [],
        'files': [],
    }
    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(args.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)

    info = tracking_output['info']

    proc_context = process_context.ProcessContext(
        name='watch.tasks.depthPCD.score_tracks', type='process',
        config=jsonified_config,
        track_emissions=False,
    )
    proc_context.start()
    info.append(proc_context.obj)

    images = kwcoco.CocoDataset.coerce(args.images_kwcoco)
    coco_dset = score_tracks(coco_dset, images, args.threshold, model_fpath)

    proc_context.stop()
    out_kwcoco = args.out_kwcoco
    if out_kwcoco is not None:
        coco_dset = coco_dset.reroot(absolute=True, check=False)
        # Add tracking audit data to the kwcoco file
        coco_info = coco_dset.dataset.get('info', [])
        coco_info.append(proc_context.obj)
        coco_dset.fpath = out_kwcoco
        ub.Path(out_kwcoco).parent.ensuredir()
        print(f'write to coco_dset.fpath={coco_dset.fpath}')
        coco_dset.dump(out_kwcoco, indent=2)

        # Convert kwcoco to sites
    import safer
    verbose = 1

    if args.out_sites_dir is not None:

        sites_dir = ub.Path(args.out_sites_dir).ensuredir()
        # Also do this in BAS mode
        sites = convert_kwcoco_to_iarpa(coco_dset,
                                        default_region_id=args.region_id,
                                        as_summary=False)
        print(f'{len(sites)=}')
        # write sites to disk
        site_fpaths = []
        for site in ub.ProgIter(sites, desc='writing sites', verbose=verbose):
            site_props = site['features'][0]['properties']
            assert site_props['type'] == 'site'
            site_fpath = sites_dir / (site_props['site_id'] + '.geojson')
            site_fpaths.append(os.fspath(site_fpath))

            with safer.open(site_fpath, 'w', temp_file=not ub.WIN32) as f:
                geojson.dump(site, f, indent=2)

    if args.out_sites_fpath is not None:
        site_tracking_output = tracking_output.copy()
        site_tracking_output['files'] = site_fpaths
        out_sites_fpath = ub.Path(args.out_sites_fpath)
        print(f'Write tracked site result to {out_sites_fpath}')
        with safer.open(out_sites_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(site_tracking_output, file, indent='    ')

    # Convert kwcoco to sites summaries
    if args.out_site_summaries_dir is not None:
        sites = convert_kwcoco_to_iarpa(coco_dset,
                                        default_region_id=args.region_id,
                                        as_summary=True)
        print(f'{len(sites)=}')
        site_summary_dir = ub.Path(args.out_site_summaries_dir).ensuredir()
        # write sites to region models on disk
        groups = ub.group_items(sites, lambda site: site['properties'].pop('region_id'))

        site_summary_fpaths = []
        for region_id, site_summaries in groups.items():

            region_fpath = site_summary_dir / (region_id + '.geojson')
            if args.append_mode and region_fpath.is_file():
                with open(region_fpath, 'r') as f:
                    region = geojson.load(f)
                if verbose:
                    print(f'writing to existing region {region_fpath}')
            else:
                region = geojson.FeatureCollection(
                    [create_region_header(region_id, site_summaries)])
                if verbose:
                    print(f'writing to new region {region_fpath}')
            for site_summary in site_summaries:
                assert site_summary['properties']['type'] == 'site_summary'
                region['features'].append(site_summary)

            site_summary_fpaths.append(os.fspath(region_fpath))
            with safer.open(region_fpath, 'w', temp_file=not ub.WIN32) as f:
                geojson.dump(region, f, indent=2)

    if args.out_site_summaries_fpath is not None:
        site_summary_tracking_output = tracking_output.copy()
        site_summary_tracking_output['files'] = site_summary_fpaths
        out_site_summaries_fpath = ub.Path(args.out_site_summaries_fpath)
        out_site_summaries_fpath.parent.ensuredir()
        print(f'Write tracked site summary result to {out_site_summaries_fpath}')
        with safer.open(out_site_summaries_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(site_summary_tracking_output, file, indent='    ')


r'''
Ignore:

    python -m watch.tasks.depthPCD.score_tracks /media/barcelona/Drop6/bas_baseline/polyb.kwcoco.zip
            --images /media/barcelona/Drop6/valT.kwcoco.zip
            --out_site_summaries_fpath "/media/barcelona/Drop6/tronexperiments/debug/site_summaries_manifest.json"
            --out_site_summaries_dir "/media/barcelona/Drop6/tronexperiments/debug/site_summaries"
            --out_sites_fpath "/media/barcelona/Drop6/tronexperiments/debug/sites_manifest.json"
            --out_sites_dir "/media/barcelona/Drop6/tronexperiments/debug/sites"
            --out_kwcoco "some file with filtered poly if you need"
            --threshold 0.3 (default)


Example:

    ### Run BAS and then run SV on top of it.

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    BAS_MODEL_FPATH=$DVC_EXPT_DPATH/models/fusion/Drop6-MeanYear10GSD-V2/packages/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47/Drop6_TCombo1Year_BAS_10GSD_V2_landcover_split6_V47_epoch47_step3026.pt

    # Predict BAS Heatmaps
    python -m watch.tasks.fusion.predict \
        --package_fpath="$BAS_MODEL_FPATH" \
        --test_dataset=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns-KR_R002_I2L.kwcoco.zip \
        --pred_dataset=$DVC_EXPT_DPATH/_test_dzyne_sv/pred_heatmaps.kwcoco.zip \
        --chip_overlap="0.3" \
        --chip_dims="196,196" \
        --time_span="auto" \
        --fixed_resolution="10GSD" \
        --time_sampling="soft4" \
        --drop_unused_frames="True"  \
        --num_workers="4" \
        --devices="0," \
        --batch_size="1" \
        --with_saliency="True" \
        --with_class="False" \
        --with_change="False"

    # Convert Heatmaps to Polygons
    python -m watch.cli.run_tracker \
        --in_file "$DVC_EXPT_DPATH/_test_dzyne_sv/pred_heatmaps.kwcoco.zip" \
        --default_track_fn saliency_heatmaps \
        --track_kwargs '{
            \"agg_fn\": \"probs\",
            \"thresh\": 0.4,
            \"time_thresh\": 0.8,
            \"inner_window_size\": \"1y\",
            \"inner_agg_fn\": \"mean\",
            \"norm_ord\": \"inf\",
            \"resolution\": \"10GSD\",
            \"moving_window_size\": null,
            \"poly_merge_method\": \"v2\",
            \"polygon_simplify_tolerance\": 1,
            \"min_area_square_meters\": 7200,
            \"max_area_square_meters\": 8000000
        }' \
        --clear_annots=True \
        --site_summary 'None' \
        --boundary_region $DVC_DATA_DPATH/annotations/drop6/region_models \
        --out_site_summaries_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/site_summaries_manifest.json" \
        --out_site_summaries_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/site_summaries" \
        --out_sites_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/sites_manifest.json" \
        --out_sites_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/sites" \
        --out_kwcoco "$DVC_EXPT_DPATH/_test_dzyne_sv/poly.kwcoco.zip"

    # Score the Initial Predictions
    python -m watch.cli.run_metrics_framework \
        --merge=True \
        --name "todo" \
        --true_site_dpath "$DVC_DATA_DPATH/annotations/drop6/site_models" \
        --true_region_dpath "$DVC_DATA_DPATH/annotations/drop6/region_models" \
        --pred_sites "$DVC_EXPT_DPATH/_test_dzyne_sv/sites_manifest.json" \
        --tmp_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/tmp" \
        --out_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before" \
        --merge_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/poly_eval_before.json"

    # Run the Site Validation Filter
    python -m watch.tasks.depthPCD.score_tracks \
        --in_file $DVC_EXPT_DPATH/_test_dzyne_sv/poly.kwcoco.zip \
        --images_kwcoco $DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.json \
        --model_fpath $DVC_EXPT_DPATH/models/depthPCD/basicModel2.h5 \
        --out_site_summaries_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_site_summaries_manifest.json" \
        --out_site_summaries_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_site_summaries" \
        --out_sites_fpath  "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites_manifest.json" \
        --out_sites_dir  "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites" \
        --out_kwcoco "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_poly.kwcoco.zip" \
        --threshold 0.4

    # Score the Filtered Predictions
    python -m watch.cli.run_metrics_framework \
        --merge=True \
        --name "todo" \
        --true_site_dpath "$DVC_DATA_DPATH/annotations/drop6/site_models" \
        --true_region_dpath "$DVC_DATA_DPATH/annotations/drop6/region_models" \
        --pred_sites "$DVC_EXPT_DPATH/_test_dzyne_sv/filtered_sites_manifest.json" \
        --tmp_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/tmp" \
        --out_dir "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after" \
        --merge_fpath "$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/poly_eval_after.json"

    python -c "if 1:
        import pandas as pd
        import rich
        import json
        import ubelt as ub
        text1 = ub.Path('$DVC_EXPT_DPATH/_test_dzyne_sv/eval_before/poly_eval_before.json').read_text()
        text2 = ub.Path('$DVC_EXPT_DPATH/_test_dzyne_sv/eval_after/poly_eval_after.json').read_text()
        data1 = json.loads(text1)
        data2 = json.loads(text2)
        df1 = pd.read_json(json.dumps(data1['best_bas_rows']), orient='table')
        df2 = pd.read_json(json.dumps(data2['best_bas_rows']), orient='table')
        print('BEFORE:')
        rich.print(df1)
        print('After:')
        rich.print(df2)
    "
'''
if __name__ == '__main__':
    main()



import kwarray
import kwimage
import numpy as np
import torch
# import matplotlib.pyplot as plt
# from kwcoco.channel_spec import ChannelSpec  # NOQA
from functools import partial
from netharn.data.batch_samplers import PatchedBatchSampler
from netharn.data.data_containers import ItemContainer
from netharn.data.data_containers import BatchContainer
from netharn.data.data_containers import container_collate
from netharn.data.batch_samplers import PatchedRandomSampler
from netharn.data.batch_samplers import SubsetSampler


class SequenceDataset(torch.utils.data.Dataset):
    def __init__(self, sampler, window_dims, input_dims=None, channels=None,
                 rng=None, window_overlap=0.0):
        super().__init__()
        if input_dims is None:
            input_dims = window_dims[-2:]
        # print(sampler)
        self.rng = kwarray.ensure_rng(rng)
        self.sampler = sampler
        self.window_dims = window_dims
        self.input_dims = input_dims
        self.classes = self.sampler.classes
        self.channels = channels
        # Build a simple space-time-grid
        sample_grid_spec = {
            'task': 'video_detection',
            'window_dims': window_dims,
            'window_overlap': window_overlap,
        }
        # print(sample_grid_spec)
        self.sample_grid = sampler.new_sample_grid(**sample_grid_spec)
        # print(len(self.sample_grid['positives']))

    def __len__(self):
        return len(self.sample_grid['positives'])
        # return len(self.sample_grid['negatives'])

    def __getitem__(self, index):

        tr = self.sample_grid['positives'][index]
        # tr = self.sample_grid['negatives'][index]
        if self.channels:
            tr['channels'] = self.channels

        sampler = self.sampler

        # import pdb
        # pdb.set_trace()
        # print(f"frame min: {frame.min()}, frame max: {frame.max()}")
        sample = sampler.load_sample(tr, with_annots="segmentation")
        # print(sample.keys())
        # print(sample['annots'].keys())
        # print(sample['annots']['rel_ssegs'])
        # print(sample['annots']['frame_dets'])
        # rel_ssegs = sample['annots']['rel_ssegs']
        # print(rel_ssegs.data)
        # seg = kwimage.Segmentation.coerce(rel_ssegs.data)
        # plt.imshow(rel_ssegs)
        # plt.show()
        # Access the sampled image and annotation data
        raw_frame_list = sample['im']
        raw_det_list = sample['annots']['frame_dets']
        # print(sample)
        # Break data down on a per-frame basis so we can apply image-based
        # augmentations.
        frame_ims = []
        frame_masks = []
        for raw_frame, raw_dets in zip(raw_frame_list, raw_det_list):
            frame = raw_frame.astype(np.float32)
            dets = raw_dets
            input_dsize = self.input_dims[-2:][::-1]
            # Resize the sampled window to the target space for the network
            frame, info = kwimage.imresize(frame, dsize=input_dsize,
                                           interpolation='linear',
                                           return_info=True)
            # Remember to apply any transform to the dets as well
            dets = dets.scale(info['scale'])
            # print(info)
            dets = dets.translate(info['offset'])
            # print(frame.shape[0:2])
            frame_mask = np.full(frame.shape[0:2], dtype=np.int32, fill_value=-1)
            ann_polys = dets.data['segmentations'].to_polygon_list()
            # print(ann_polys)
            ann_aids = dets.data['aids']
            ann_cids = dets.data['cids']
            # print(ann_cids)
            for poly, aid, cid in zip(ann_polys, ann_aids, ann_cids):
                cidx = self.classes.id_to_idx[cid]
                poly.fill(frame_mask, value=cidx)

            # ensure channel dim is not squeezed
            frame = kwarray.atleast_nd(frame, 3)

            # fig = plt.figure()
            # ax1 = fig.add_subplot(1,2,1)
            # ax2 = fig.add_subplot(1,2,2)
            # ax1.imshow(frame[:,:,:3])
            # ax2.imshow(frame_mask)
            # plt.show()

            frame_masks.append(frame_mask)
            frame_ims.append(frame)

        # Perpare data for torch
        frame_data = np.concatenate([f[None, ...] for f in frame_ims], axis=0)
        class_masks = np.concatenate([m[None, ...] for m in frame_masks], axis=0)

        cthw_im = frame_data.transpose(3, 0, 1, 2)

        inputs = {
            'im': ItemContainer(torch.from_numpy(cthw_im), stack=True),
        }
        label = {
            'class_masks': ItemContainer(
                torch.from_numpy(class_masks), stack=False, cpu_only=True),
        }

        item = {
            'inputs': inputs,
            'label': label,
            'tr': ItemContainer(sample['tr'], stack=False),
        }
        return item

    def make_loader(self, batch_size=16, num_workers=0, shuffle=False,
                    pin_memory=True, drop_last=True, multiscale=False,
                    balance=False, num_batches='auto', xpu=None):
        """
        Create a loader for this dataset with custom sampling logic and
        container collation.
        """
        if len(self) == 0:
            raise ValueError('must have some data')

        if balance:
            # Can use information in self.sample_grid to balance over
            # categories.
            raise NotImplementedError

        # The case where where replacement is not allowed
        if num_batches == 'auto':
            num_samples = None
        else:
            num_samples = num_batches * batch_size

        if shuffle:
            item_sampler = PatchedRandomSampler(self, num_samples=num_samples)
        else:
            if num_samples is None:
                item_sampler = torch.utils.data.sampler.SequentialSampler(self)
            else:
                stats_idxs = (np.arange(num_samples) % len(self))
                item_sampler = SubsetSampler(stats_idxs)

        if num_samples is not None:
            # If num_batches is too big, error
            if num_samples > len(self):
                raise IndexError(
                    'num_batches={} and batch_size={} causes '
                    'num_samples={} to be greater than the number '
                    'of data items {}. Try setting num_batches=auto?'.format(
                        num_batches, batch_size, num_samples, len(self)))

        batch_sampler = PatchedBatchSampler(
            item_sampler, batch_size=batch_size, drop_last=drop_last,
            num_batches=num_batches)

        if xpu is None:
            num_devices = 1
        else:
            num_devices = len(xpu.devices)

        collate_fn = partial(container_collate, num_devices=num_devices)

        loader = torch.utils.data.DataLoader(
            self, batch_sampler=batch_sampler,
            collate_fn=collate_fn, num_workers=num_workers,
            pin_memory=pin_memory, worker_init_fn=worker_init_fn)
        return loader


def worker_init_fn(worker_id):
    worker_info = torch.utils.data.get_worker_info()
    self = worker_info.dataset

    if hasattr(self.sampler.dset, 'connect'):
        # Reconnect to the backend if we are using SQL
        self.sampler.dset.connect(readonly=True)


def decollate_batch(batch):
    """
    Breakup a collated batch of BatchContainers back into ItemContainers
    """
    import ubelt as ub
    walker = ub.IndexableWalker(batch)
    decollated_dict = ub.AutoDict()
    decollated_walker = ub.IndexableWalker(decollated_dict)
    for path, batch_val in walker:
        if isinstance(batch_val, BatchContainer):
            for bx, item_val in enumerate(ub.flatten(batch_val.data)):
                decollated_walker[[bx] + path] = ItemContainer(item_val)
    decollated = list(decollated_dict.to_dict().values())
    return decollated


# def __notes__():
#     """
#         >>> #
#         >>> for img in dset.imgs.values():
#         >>>     chan = img.get('channels', None)
#         >>>     print('img_chan = {!r}'.format(chan))
#         >>>     for aux in img.get('auxiliary', []):
#         >>>         chan = aux.get('channels', None)
#         >>>         print('aux_chan = {!r}'.format(chan))
#     """



import torch
from datetime import datetime

from watch.tasks.rutgers_material_seg_v2.datasets.image_dataset import ImageDataset
from watch.tasks.rutgers_material_seg_v2.datasets.material_pixel_dataset import MaterialPixelDataset

DATASETS = {'material_pixel': {'training': MaterialPixelDataset, 'evaluation': ImageDataset}}


def build_dataset(dataset_name, dataset_path, split, **kwargs):
    if split in ['train', 'valid']:
        split_type = 'training'
    elif split in ['test', 'eval']:
        split_type = 'evaluation'
    else:
        raise NotImplementedError

    try:
        dataset_type = DATASETS[dataset_name]
    except KeyError:
        raise NotImplementedError(f'Dataset "{dataset_name}" has not been implemented in the build_dataset function.')

    try:
        dataset_func = dataset_type[split_type]
    except KeyError:
        raise NotImplementedError(
            f'The {split_type} dataset for "{dataset_name}" has not been implemented for build_dataset function.')

    dataset = dataset_func(dataset_path, split, **kwargs)

    return dataset


def custom_collate_fn(dset_outputs):
    out_data = {}
    keys = dset_outputs[0].keys()
    for key in keys:
        # Return a list of items.
        out = []
        for output in dset_outputs:
            if output[key] is None:
                pass

            if type(output[key]) == datetime:
                out.append(output[key])
            else:
                out.append(torch.tensor(output[key]))

        # Handle edge case variables from output of dataset class.
        if type(dset_outputs[0][key]) == datetime:
            pass
        else:
            out = torch.stack(out, 0)
        out_data[key] = out

    return out_data



"""
This is step 2/4 in predict.py and the step that runs pycold


SeeAlso:

    predict.py

    prepare_kwcoco.py

    tile_processing_kwcoco.py *

    export_cold_result_kwcoco.py

    assemble_cold_result_kwcoco.py

This script is for running COLD algorithm with kwcoco dataset.
See original code: ~/code/pycold/src/python/pycold/imagetool/tile_processing.py
"""

import os
import time
import json
import pandas as pd
import numpy as np
from datetime import datetime as datetime_cls
from pytz import timezone
from scipy.stats import chi2
import ubelt as ub  # NOQA
import pycold  # NOQA
from pycold import cold_detect
from pycold.utils import get_rowcol_intile, get_doy, assemble_cmmaps
from pycold.ob_analyst import ObjectAnalystHPC
import scriptconfig as scfg
from pathlib import Path

try:
    from xdev import profile
except ImportError:
    from ubelt import identity as profile


class TileProcessingKwcocoConfig(scfg.DataConfig):
    """
    The docstring will be the description in the CLI help
    """
    rank = scfg.Value(None, help='rank id')
    n_cores = scfg.Value(None, help='total cores assigned (parent context, not workers used by this process)')
    stack_path = scfg.Value(None, help='directory of stacked data')
    reccg_path = scfg.Value(None, help='directory where cold record will be saved')
    meta_fpath = scfg.Value(None, help='json file created by prepare_kwcoco.py')
    method = scfg.Value('COLD', choices=['COLD', 'HybridCOLD', 'OBCOLD'], help='type of COLD algorithms, e.g., COLD, HybridCOLD, OBCOLD')
    b_c2 = scfg.Value(True, help='indicate if it is c2 or not')
    prob = scfg.Value(0.99, help='change probability of chi-distribution, e.g., 0.99')
    conse = scfg.Value(6, help='consecutive observation to confirm change, e.g., 6')
    cm_interval = scfg.Value(60, help='CM output inverval, e.g., 60')


@profile
def tile_process_main(cmdline=1, **kwargs):
    """
    Args:
        n_cores (type=int): _description_
        stack_path (_type_): _description_
        reccg_path (_type_): _description_
        method (_type_): _description_
        year_lowbound (_type_): _description_
        year_highbound (_type_): _description_
        b_c2 (_type_): _description_
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m watch.tasks.cold.tile_processing_kwcoco --help
        TEST_COLD=1 xdoctest -m watch.tasks.cold.tile_processing_kwcoco tile_process_main

    Example:
    >>> # xdoctest: +REQUIRES(env:TEST_COLD)
    >>> from watch.tasks.cold.tile_processing_kwcoco import tile_process_main
    >>> from watch.tasks.cold.tile_processing_kwcoco import *
    >>> kwargs= dict(
    >>>    rank = 1,
    >>>    n_cores = 1,
    >>>    stack_path = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/kwcoco/stacked/KR_R001'),
    >>>    reccg_path = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/kwcoco/reccg/KR_R001'),
    >>>    meta_fpath = '/gpfs/scratchfs1/zhz18039/jws18003/kwcoco/stacked/KR_R001/block_x10_y1/crop_20140115T020000Z_N37.643680E128.649453_N37.683356E128.734073_L8_0.json',
    >>>    method = 'COLD',
    >>>    b_c2 = True,
    >>>    prob = 0.99,
    >>>    conse = 6,
    >>>    cm_interval = 60,
    >>> )
    >>> cmdline=0
    >>> tile_process_main(cmdline, **kwargs)
    """
    # Hacky way to pass in progress manager
    pman = kwargs.pop('pman', None)

    # setting config
    config_in = TileProcessingKwcocoConfig.cli(cmdline=cmdline, data=kwargs)
    rank = config_in['rank']
    n_cores = config_in['n_cores']
    stack_path = Path(config_in['stack_path'])
    reccg_path = Path(config_in['reccg_path'])
    meta_fpath = Path(config_in['meta_fpath'])
    method = config_in['method']
    b_c2 = config_in['b_c2']
    prob = config_in['prob']
    conse = config_in['conse']
    cm_output_interval = config_in['cm_interval']

    config = json.loads(meta_fpath.read_text())
    n_cols = config['padded_n_cols']
    n_rows = config['padded_n_rows']
    n_block_x = config['n_block_x']
    n_block_y = config['n_block_y']
    block_width = int(n_cols / n_block_x)  # width of a block
    block_height = int(n_rows / n_block_y)  # height of a block
    year_lowbound = None
    year_highbound = None

    tz = timezone('US/Eastern')
    start_time = datetime_cls.now(tz)

    # Define year_low_ordinal and year_high_ordinal to filter year for COLD processing
    if year_lowbound is None:
        year_lowbound = 0
    else:
        year_lowbound = pd.Timestamp.toordinal(datetime_cls(int(year_lowbound), 1, 1))

    if year_highbound is None:
        year_highbound = 0
    else:
        year_highbound = pd.Timestamp.toordinal(datetime_cls(int(year_highbound + 1), 1, 1))

    if (n_cols % block_width != 0) or (n_rows % block_height != 0):
        print('padded_n_cols, padded_n_rows must be divisible respectively by block_width, block_height! Please double '
              'check your config yaml')
        exit()

    # set up additional parameters for obcold
    if method == 'OBCOLD':
        # we need read 'global starting date' to save CM which will be only used for ob-cold
        try:
            starting_date, n_cm_maps = reading_start_dates_nmaps(stack_path, year_lowbound, year_highbound,
                                                                 cm_output_interval)
            year_lowbound = pd.Timestamp.fromordinal(starting_date).year
            year_highbound = pd.Timestamp.fromordinal(
                starting_date + (n_cm_maps - 1) * cm_output_interval).year
        except IOError:
            print(f"reading start dates errors: {start_time.strftime('%Y-%m-%d %H:%M:%S')}")
            exit()

    # logging and folder preparation
    # if rank == 0:
    reccg_path.mkdir(parents=True, exist_ok=True)
    if method == 'OBCOLD':
        cm_maps_fpath = reccg_path / 'cm_maps'
        cm_maps_fpath.mkdir(parents=True, exist_ok=True)
    print(f"The per-pixel time series processing begins: {start_time.strftime('%Y-%m-%d %H:%M:%S')}")

    if not stack_path.exists():
        print("Failed to locate stack folders. The program ends: "
                f"{datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')}")
        return

    #########################################################################
    #                        per-pixel COLD procedure                       #
    #########################################################################
    threshold = chi2.ppf(prob, 5)

    nblock_eachcore = int(np.ceil(n_block_x * n_block_y * 1.0 / n_cores))
    i_iter = range(nblock_eachcore + 1)
    if pman is not None:
        i_iter = pman.progiter(i_iter, desc=f'Process Tile \\[rank {rank}]',
                               total=nblock_eachcore, transient=True)
    for i in i_iter:
        block_id = n_cores * i + rank  # started from 1, i.e., rank, rank + n_cores, rank + 2 * n_cores
        if block_id > n_block_x * n_block_y :
            break
        block_y = int((block_id - 1) / n_block_x ) + 1  # note that block_x and block_y start from 1
        block_x = int((block_id - 1) % n_block_x ) + 1

        finished_fpath = reccg_path / f'COLD_block{block_id}_finished.txt'
        if finished_fpath.exists():
            now = datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')
            print(f"Per-pixel COLD processing is finished for block_x{block_x}_y{block_y} ({now})")
            continue
        img_tstack, img_dates_sorted = get_stack_date(block_x, block_y, stack_path, year_lowbound,
                                                      year_highbound)

        # Define empty list
        result_collect = []
        date_collect = []
        CM_collect = []

        if img_tstack is None:  # empty block
            if method == 'OBCOLD':
                for pos in range(block_width * block_height):
                    CM_collect.append(np.full(n_cm_maps, -9999, dtype=np.short))
                    date_collect.append(np.full(n_cm_maps, -9999, dtype=np.short))
                np.save(reccg_path / f'CM_date_x{block_x}_y{block_y}.npy', np.hstack(date_collect))
                np.save(reccg_path / f'CM_x{block_x}_y{block_y}.npy', np.hstack(CM_collect))

        else:
            # start looping every pixel in the block
            if method == "COLD" or method == "HybridCOLD" or method == "OBCOLD":
                for pos in range(block_width * block_height):
                    original_row, original_col = get_rowcol_intile(pos, block_width,
                                                                   block_height, block_x, block_y)
                    try:
                        if method == 'OBCOLD':
                            [cold_result, CM, CM_date] = cold_detect(img_dates_sorted,
                                                                     img_tstack[pos, 0, :].astype(np.int64),
                                                                     img_tstack[pos, 1, :].astype(np.int64),
                                                                     img_tstack[pos, 2, :].astype(np.int64),
                                                                     img_tstack[pos, 3, :].astype(np.int64),
                                                                     img_tstack[pos, 4, :].astype(np.int64),
                                                                     img_tstack[pos, 5, :].astype(np.int64),
                                                                     img_tstack[pos, 6, :].astype(np.int64),
                                                                     img_tstack[pos, 7, :].astype(np.int64),
                                                                     pos=n_cols * (original_row - 1) + original_col,
                                                                     conse=conse,
                                                                     starting_date=starting_date,
                                                                     n_cm=n_cm_maps, b_c2=b_c2,
                                                                     cm_output_interval=cm_output_interval,
                                                                     b_output_cm=True)
                        else:
                            cold_result = cold_detect(img_dates_sorted,
                                                      img_tstack[pos, 0, :].astype(np.int64),
                                                      img_tstack[pos, 1, :].astype(np.int64),
                                                      img_tstack[pos, 2, :].astype(np.int64),
                                                      img_tstack[pos, 3, :].astype(np.int64),
                                                      img_tstack[pos, 4, :].astype(np.int64),
                                                      img_tstack[pos, 5, :].astype(np.int64),
                                                      img_tstack[pos, 6, :].astype(np.int64),
                                                      img_tstack[pos, 7, :].astype(np.int64),
                                                      t_cg=threshold,
                                                      conse=conse, b_c2=b_c2,
                                                      pos=n_cols * (original_row - 1) + original_col)

                    except RuntimeError:
                        now = datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')
                        print(f"COLD fails at original_row {original_row}, original_col {original_col} ({now})")
                    except Exception:
                        if method == 'OBCOLD':
                            CM = np.full(n_cm_maps, -9999, dtype=np.short)
                            CM_date = np.full(n_cm_maps, -9999, dtype=np.short)
                    else:
                        result_collect.append(cold_result)
                    finally:
                        if method == 'OBCOLD':
                            CM_collect.append(CM)
                            date_collect.append(CM_date)

                # save the dataset
                if len(result_collect) == 0:
                    with open(reccg_path / f'record_change_x{block_x}_y{block_y}_cold.status.json', 'w') as f:
                        json.dump({'status': 'failed', 'output': 'invalid'}, f)
                if len(result_collect) > 0:
                    if method == 'HybridCOLD':
                        np.save(reccg_path / f'record_change_x{block_x}_y{block_y}_hybridcold.npy',
                                np.hstack(result_collect))
                    elif method == 'COLD':
                        np.save(reccg_path / f'record_change_x{block_x}_y{block_y}_cold.npy',
                                np.hstack(result_collect))
                    with open(reccg_path / f'record_change_x{block_x}_y{block_y}_cold.status.json', 'w') as f:
                        json.dump({'status': 'completed', 'output': 'valid'}, f)
                if method == 'OBCOLD':
                    np.save(reccg_path / 'CM_date_x{block_x}_y{block_y}.npy', np.hstack(date_collect))
                    np.save(reccg_path / 'CM_x{block_x}_y{block_y}.npy', np.hstack(CM_collect))

            with open(reccg_path / f'COLD_block{block_id}_finished.txt', 'w'):
                pass

            now = datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')
            print(f"Per-pixel COLD processing is finished for block_x{block_x}_y{block_y} ({now})")

    # wait for all cores to be finished
    if method == 'OBCOLD':
        while not is_finished_cold_blockfinished(reccg_path, n_block_x * n_block_y):
            time.sleep(30)

    # if rank == 1:
    #     cold_timepoint = datetime_cls.now(tz)
    # for i in os.listdir(reccg_path):
    #     if i.endswith('.txt'):
    #         os.remove(reccg_path / i)

    #################################################################################
    #                        the below is object-based process                      #
    #################################################################################
    if method == 'OBCOLD':
        # if seedmap_path is None:
        ob_analyst = ObjectAnalystHPC(config, starting_date=starting_date, stack_path=stack_path,
                                          result_path=reccg_path)
        # else:
        #     pyclassifier = PyClassifierHPC(config, record_path=reccg_path, year_lowbound=year_lowbound,
        #                                    year_uppbound=year_highbound, seedmap_path=seedmap_path)
        #     ob_analyst = ObjectAnalystHPC(config, starting_date=starting_date, stack_path=stack_path,
        #                                   result_path=reccg_path, thematic_path=reccg_path / 'feature_maps')
        if rank == 1:
            # need to create folders first
            # if seedmap_path is not None:
            #     pyclassifier.hpc_preparation()
            ob_analyst.hpc_preparation()

        #########################################################################
        #                        reorganize cm snapshots                        #
        #########################################################################

        if not is_finished_assemble_cmmaps(reccg_path / 'cm_maps', n_cm_maps,
                                           starting_date, cm_output_interval):
            if rank == 1:
                assemble_cmmaps(config, reccg_path, reccg_path / 'cm_maps', starting_date, n_cm_maps, 'CM',
                                clean=False)
            elif rank == 2:
                assemble_cmmaps(config, reccg_path, reccg_path / 'cm_maps', starting_date, n_cm_maps,
                                'CM_date',
                                clean=False)

            while not is_finished_assemble_cmmaps(reccg_path / 'cm_maps', n_cm_maps,
                                                  starting_date, cm_output_interval):
                time.sleep(15)

        #########################################################################
        #                      producing classification maps                    #
        #########################################################################
        # if seedmap_path is not None:  # we used thematic info
        #     if not pyclassifier.is_finished_step4_assemble():
        #         if rank == 1:
        #             print("Starts predicting features: {}".format(datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')))
        #         for i in range(nblock_eachcore):
        #             if n_cores * i + rank > config['n_block_x'] * config['n_block_y']:
        #                 break
        #             pyclassifier.step1_feature_generation(block_id=n_cores * i + rank)

        #         if rank == 1:  # serial mode for producing rf
        #             pyclassifier.step2_train_rf()
        #             print("Training rf ends: {}".format(datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')))

        #         for i in range(nblock_eachcore):
        #             if n_cores * i + rank > config['n_block_x'] * config['n_block_y']:
        #                 break
        #             pyclassifier.step3_classification(block_id=n_cores * i + rank)

        #         if rank == 1:  # serial mode for assemble
        #             pyclassifier.step4_assemble()
        #     while not pyclassifier.is_finished_step4_assemble():
        #         time.sleep(15)
        #     if rank == 1:
        #         print("Assemble classification map ends: {}".format(datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')))
        #########################################################################
        #                      object-based image analysis                      #
        #########################################################################
        if not ob_analyst.is_finished_object_analysis(np.arange(starting_date,
                                                                starting_date + cm_output_interval * n_cm_maps,
                                                                cm_output_interval)):
            n_map_percore = int(np.ceil(n_cm_maps / n_cores))
            max_date = starting_date + (n_cm_maps - 1) * cm_output_interval
            for i in range(n_map_percore):
                if starting_date + (rank - 1 + i * n_cores) * cm_output_interval > max_date:
                    break
                date = starting_date + (rank - 1 + i * n_cores) * cm_output_interval
                ob_analyst.obia_execute(date)

            while not ob_analyst.is_finished_object_analysis(np.arange(starting_date,
                                                                       starting_date + cm_output_interval * n_cm_maps,
                                                                       cm_output_interval)):
                time.sleep(15)
        if rank == 1:
            print(f"OBIA ends: {datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')}")

        #########################################################################
        #                        reconstruct change records                     #
        #########################################################################
        for i in range(nblock_eachcore):
            block_id = n_cores * i + rank  # started from 1, i.e., rank, rank + n_cores, rank + 2 * n_cores
            if block_id > n_block_x * n_block_y:
                break
            block_y = int((block_id - 1) / n_block_y) + 1  # note that block_x and block_y start from 1
            block_x = int((block_id - 1) % n_block_x) + 1
            img_tstack, img_dates_sorted = get_stack_date(block_x, block_y, stack_path)
            result_collect = ob_analyst.reconstruct_reccg(block_id=block_id,
                                                          img_stack=img_tstack,
                                                          img_dates_sorted=img_dates_sorted)
            ob_analyst.save_obcoldrecords(block_id=block_id, result_collect=result_collect)

    # if rank == 1:
    #     # tile_based report
    log = {
        'algorithm': method,
        'prob': prob,
        'conse': conse,
    }

    log_fpath = reccg_path / 'log.json'
    log_fpath.write_text(json.dumps(log))

    # if method == 'OBCOLD':
    #     tileprocessing_report(reccg_path / 'tile_processing_report.log',
    #                           stack_path, pycold.__version__, method, config, start_time, cold_timepoint, tz,
    #                           n_cores, starting_date, n_cm_maps, year_lowbound, year_highbound)
    # else:
    #     tileprocessing_report(reccg_path / 'tile_processing_report.log', stack_path, pycold.__version__,
    #                           method, config, start_time, cold_timepoint, tz, n_cores)
    # print("The whole procedure finished: {}".format(datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')))


# def tileprocessing_report(result_log_path, stack_path, version, algorithm, config, startpoint, cold_timepoint, tz,
#                           n_cores, starting_date=0, n_cm_maps=0, year_lowbound=0, year_highbound=0):
#     """
#     output tile-based processing report
#     Parameters
#     ----------
#     result_log_path: string
#         outputted log path
#     stack_path: string
#         stack path
#     version: string
#     algorithm: string
#     config: dictionary structure
#     startpoint: a time point, when the program starts
#     tz: string, time zone
#     n_cores: the core number used
#     starting_date: the first date of the total dataset
#     n_cm_maps: the number of snapshots
#     year_lowbound: the low bound of year range
#     year_highbound: the upper bound of year range
#     Returns
#     -------
#     """
#     endpoint = datetime_cls.now(tz)
#     file = open(result_log_path, "w")
#     file.write("PYCOLD V{} \n".format(version))
#     file.write("Author: Su Ye(remoteseningsuy@gmail.com)\n")
#     file.write("Algorithm: {} \n".format(algorithm))
#     file.write("Starting_time: {}\n".format(startpoint.strftime('%Y-%m-%d %H:%M:%S')))
#     file.write("Change probability threshold: {}\n".format(change_probability))
#     file.write("Conse: {}\n".format(conse))
#     file.write("stack_path: {}\n".format(stack_path))
#     file.write("The number of requested cores: {}\n".format(n_cores))
#     file.write("The program starts at {}\n".format(startpoint.strftime('%Y-%m-%d %H:%M:%S')))
#     file.write("The COLD ends at {}\n".format(cold_timepoint.strftime('%Y-%m-%d %H:%M:%S')))
#     file.write("The program ends at {}\n".format(endpoint.strftime('%Y-%m-%d %H:%M:%S')))
#     file.write("The program lasts for {:.2f}mins\n".format((endpoint - startpoint) / datetime_cls.timedelta(minutes=1)))
#     file.close()

@profile
def is_finished_cold_blockfinished(reccg_path, nblocks):
    """
    check if the COLD algorithm finishes all blocks
    Parameters
    ----------
    reccg_path: the path that save COLD results
    nblocks: the block number

    Returns: boolean
    -------
        True -> all block finished
    """
    for n in range(nblocks):
        fpath = reccg_path / f'COLD_block{n + 1}_finished.txt'
        if not fpath.exists():
            return False
    return True


@profile
def get_stack_date(block_x, block_y, stack_path, year_lowbound=0, year_highbound=0, nband=8):
    """
    :param block_x: block id at x axis
    :param block_y: block id at y axis
    :param stack_path: stack path
    :param year_lowbound: ordinal data of low bounds of selection date range
    :param year_highbound: ordinal data of upper bounds of selection date range
    :return:
        img_tstack, img_dates_sorted
        img_tstack - 3-d array (block_width * block_height, nband, nimage)
    """
    block_folder = stack_path / f'block_x{block_x}_y{block_y}'
    meta_files = [m for m in os.listdir(block_folder) if m.endswith('.json')]

    # sort image files by ordinal dates
    img_dates = []
    img_files = []

    # read metadata and
    for meta in meta_files:
        config = json.loads((block_folder / meta).read_text())
        ordinal_date = config['ordinal_date']
        img_name = config['image_name'] + '.npy'
        img_dates.append(ordinal_date)
        img_files.append(img_name)

    if len(img_files) == 0:
        return None, None

    sample_np = np.load(block_folder / img_files[0])
    block_width = sample_np.shape[1]
    block_height = sample_np.shape[0]

    if year_lowbound > 0:
        year_low_ordinal = pd.Timestamp.toordinal(datetime_cls(int(year_lowbound), 1, 1))
        img_dates, img_files = zip(*filter(lambda x: x[0] >= year_low_ordinal,
                                           zip(img_dates, img_files)))
    if year_highbound > 0:
        year_high_ordinal = pd.Timestamp.toordinal(datetime_cls(int(year_highbound + 1), 1, 1))
        img_dates, img_files = zip(*filter(lambda x: x[0] < year_high_ordinal,
                                           zip(img_dates, img_files)))

    files_date_zip = sorted(zip(img_dates, img_files))
    img_files_sorted = [x[1] for x in files_date_zip]
    img_dates_sorted = np.asarray([x[0] for x in files_date_zip])
    img_tstack = np.dstack([np.load(block_folder / f).reshape(block_width * block_height, nband)
                            for f in img_files_sorted])
    return img_tstack, img_dates_sorted


#########################################################################
#                     function for OB-COLD procedure                    #
#########################################################################


@profile
def reading_start_dates_nmaps(stack_path, year_lowbound, year_highbound, cm_interval):
    """
    Parameters
    ----------
    stack_path: string
        stack_path for saving starting_last_dates.txt
    cm_interval: interval
        day interval for outputting change magnitudes
    Returns
    -------
        (starting_date, n_cm_maps)
        starting_date - starting date is the first date of the whole dataset,
        n_cm_maps - the number of change magnitudes to be outputted per pixel per band
    """
    # read starting and ending dates, note that all blocks only has one starting and last date (mainly for obcold)
    try:
        block_folder = stack_path / 'block_x1_y1'
        meta_files = [m for m in os.listdir(block_folder) if m.endswith('.json')]

        # sort image files by ordinal dates
        img_dates = []

        # read metadata and
        for meta in meta_files:
            config = json.loads((block_folder / meta).read_text())
            ordinal_date = config['ordinal_date']
            img_dates.append(ordinal_date)
        sorted_img_dates = sorted(img_dates)

        if year_lowbound > 0:
            year_low_ordinal = pd.Timestamp.toordinal(datetime_cls(int(year_lowbound), 1, 1))
            img_dates = (lambda x: x >= year_low_ordinal, img_dates)
        if year_highbound > 0:
            year_high_ordinal = pd.Timestamp.toordinal(datetime_cls(int(year_highbound + 1), 1, 1))
            img_dates = (lambda x: x < year_high_ordinal, img_dates)

    except IOError:
        raise
    else:
        starting_date = sorted_img_dates[0]
        ending_date = sorted_img_dates[-1]
        n_cm_maps = int((ending_date - starting_date + 1) / cm_interval) + 1
        return starting_date, n_cm_maps


@profile
def is_finished_assemble_cmmaps(cmmap_path, n_cm, starting_date, cm_interval):
    """
    Parameters
    ----------
    cmmap_path: the path for saving change magnitude maps
    n_cm: the number of change magnitudes outputted per pixel
    starting_date: the starting date of the whole dataset
    cm_interval: the day interval for outputting change magnitudes
    Returns: boolean
    -------
    True -> assemble finished
    """
    for count in range(n_cm):
        ordinal_date = starting_date + count * cm_interval
        year = pd.Timestamp.fromordinal(ordinal_date).year
        doy = get_doy(ordinal_date)
        cm_fpath = cmmap_path / f'CM_maps_{str(ordinal_date)}_{year}{doy}.npy'
        cm_date_fpath = cmmap_path / 'CM_date_maps_{str(ordinal_date)}_{year}{doy}.npy'
        if not cm_fpath.exists():
            return False
        if not cm_date_fpath.exists():
            return False
    return True


if __name__ == '__main__':
    tile_process_main()



import torch
import torch.nn as nn
from torchvision import models


def double_conv(in_channels, out_channels):
    return nn.Sequential(nn.Conv2d(in_channels, out_channels, 3, padding=1),
                         nn.ReLU(inplace=True),
                         nn.Conv2d(out_channels, out_channels, 3, padding=1),
                         nn.ReLU(inplace=True))


class UNetR(nn.Module):

    def __init__(self, num_outputs, num_channels=3):
        super().__init__()

        self.in_channels = num_channels
        self.out_channels = num_outputs

        resnet = models.resnet18(weights=models.ResNet18_Weights.DEFAULT)

        if num_channels == 3:
            self.firstconv = resnet.conv1
        else:
            self.firstconv = nn.Conv2d(num_channels,
                                       64,
                                       kernel_size=(7, 7),
                                       stride=(2, 2),
                                       padding=(3, 3))

        self.firstbn = resnet.bn1
        self.firstrelu = resnet.relu
        self.firstmaxpool = resnet.maxpool
        self.encoder1 = resnet.layer1
        self.encoder2 = resnet.layer2
        self.encoder3 = resnet.layer3
        self.encoder4 = resnet.layer4

        self.upsample = nn.Upsample(scale_factor=2,
                                    mode='bilinear',
                                    align_corners=True)

        self.decoder4 = double_conv(256 + 512, 256)
        self.decoder3 = double_conv(128 + 256, 128)
        self.decoder2 = double_conv(64 + 128, 64)
        self.decoder1 = double_conv(64 + 64, 64)

        self.conv_last = nn.Conv2d(64, num_outputs, 3, padding=1)

    def forward(self, x):
        x = self.firstconv(x)
        x = self.firstbn(x)
        e0 = self.firstrelu(x)

        e1 = self.encoder1(self.firstmaxpool(e0))
        e2 = self.encoder2(e1)
        e3 = self.encoder3(e2)
        e4 = self.encoder4(e3)

        x = self.upsample(e4)
        d3 = self.decoder4(torch.cat([x, e3], dim=1))

        x = self.upsample(d3)
        d2 = self.decoder3(torch.cat([x, e2], dim=1))

        x = self.upsample(d2)
        d1 = self.decoder2(torch.cat([x, e1], dim=1))

        x = self.upsample(d1)
        d0 = self.decoder1(torch.cat([x, e0], dim=1))

        x = self.upsample(d0)
        out = self.conv_last(x)

        return out



import ubelt as ub
import functools


try:
    from lark import Transformer
except ImportError:
    class Transformer:
        pass


try:
    cache = functools.cache
except AttributeError:
    cache = ub.memoize


# For common constructs see:
# https://github.com/lark-parser/lark/blob/master/lark/grammars/common.lark
RESOLUTION_GRAMMAR_PARTS = ub.codeblock(
    '''
    // Resolution parts of the grammar
    magnitude: NUMBER

    unit: WORD

    resolution: magnitude WS* unit

    %import common.NUMBER
    %import common.WS
    %import common.WORD
    ''')


RESOLVED_WINDOW_GRAMMAR = ub.codeblock(
    r'''
    // RESOLVED WINDOW GRAMMAR
    ?start: resolved_window

    window_1d_dim: NUMBER

    window_2d_dim: NUMBER WS* ("x" | ",") WS* NUMBER

    window: window_1d_dim | window_2d_dim

    resolved_window: window WS* "@" WS* resolution

    %import common.NUMBER
    %import common.WS
    ''') + '\n' + RESOLUTION_GRAMMAR_PARTS


RESOLVED_SCALAR_GRAMMAR = ub.codeblock(
    r'''
    // RESOLVED WINDOW GRAMMAR
    ?start: resolved_scalar

    resolved_scalar: NUMBER WS* "@" WS* resolution

    %import common.NUMBER
    %import common.WS
    ''') + '\n' + RESOLUTION_GRAMMAR_PARTS


def _int_or_float(x):
    try:
        return int(x)
    except Exception:
        return float(x)


class ExtendedTransformer(Transformer):
    """
    Enriches the Transformer with parse and parser classmethods which rely on a
    __grammar__ attribute
    """

    @classmethod
    @cache
    def parser(cls):
        # https://github.com/lark-parser/lark/blob/master/docs/_static/lark_cheatsheet.pdf
        import lark
        try:
            import lark_cython
            parser = lark.Lark(cls.__grammar__, start='start', parser='lalr', _plugins=lark_cython.plugins)
        except ImportError:
            parser = lark.Lark(cls.__grammar__, start='start', parser='lalr')
        return parser

    @classmethod
    @cache
    def parse(cls, text):
        parser = cls.parser()
        tree = parser.parse(text)
        transformed = cls().transform(tree)
        return transformed


class ResolvedTransformer(ExtendedTransformer):

    def magnitude(self, items):
        d = _int_or_float(items[0].value)
        return d

    def unit(self, items):
        return items[0].value

    def resolution(self, items):
        info = {
            'mag': items[0],
            'unit': items[-1],
        }
        return info


class ResolvedWindowTransformer(ResolvedTransformer):
    __grammar__ = RESOLVED_WINDOW_GRAMMAR

    def window_1d_dim(self, items):
        d1 = _int_or_float(items[0].value)
        info = (d1, d1)
        return info

    def window_2d_dim(self, items):
        d1 = _int_or_float(items[0].value)
        d2 = _int_or_float(items[-1].value)
        info = (d1, d2)
        return info

    def window(self, items):
        return items[0]

    def resolved_window(self, items):
        info = {
            'window': items[0],
            'resolution': items[-1],
        }
        return info


class ResolvedScalarTransformer(ResolvedTransformer):
    __grammar__ = RESOLVED_SCALAR_GRAMMAR

    def resolved_scalar(self, items):
        info = {
            'scalar': _int_or_float(items[0].value),
            'resolution': items[-1],
        }
        return info


class Resolved:
    """
    Base class for all resolved objects
    """

    @classmethod
    def parse(cls, data):
        if isinstance(data, str):
            return cls(**cls.__transformer__.parse(data))
        else:
            raise TypeError(type(data))

    @classmethod
    def coerce(cls, data):
        if isinstance(data, cls):
            return data
        elif isinstance(data, str):
            return cls.parse(data)
        elif isinstance(data, dict):
            return cls(**data)
        else:
            raise TypeError(type(data))


class ResolvedScalar(Resolved, ub.NiceRepr):
    """
    Example:
        >>> from watch.utils.util_resolution import *  # NOQA
        >>> resolved_scalar = "128@10GSD"
        >>> resolved_scalar = ResolvedScalar.parse(resolved_scalar)
        >>> print('resolved_scalar = {}'.format(ub.urepr(resolved_scalar, sv=1, nl=1)))
        resolved_scalar = <ResolvedScalar(128 @ {'mag': 10, 'unit': 'GSD'})>
        >>> resolved_scalar = "128  @  10  GSD"
        >>> resolved_scalar = ResolvedScalar.parse(resolved_scalar)
        >>> print('resolved_scalar = {}'.format(ub.urepr(resolved_scalar, sv=1, nl=1)))
        resolved_scalar = <ResolvedScalar(128 @ {'mag': 10, 'unit': 'GSD'})>
    """
    __transformer__ = ResolvedScalarTransformer

    def __init__(self, scalar, resolution):
        self.scalar = scalar
        self.resolution = resolution

    def __nice__(self):
        return (f'{self.scalar} @ {self.resolution}')

    def at_resolution(self, new_resolution):
        '''
        Update the resolution

        Example:
            >>> new_resolution = {'mag': 1, 'unit': 'GSD'}
            >>> self = ResolvedScalar.parse("128@10GSD")
            >>> print(self.at_resolution(new_resolution))
            >>> print(self.at_resolution({'mag': 20, 'unit': 'GSD'}))
            <ResolvedScalar(1280.0 @ {'mag': 1, 'unit': 'GSD'})>
            <ResolvedScalar(64.0 @ {'mag': 20, 'unit': 'GSD'})>
        '''
        scale_factor = self.resolution['mag'] / new_resolution['mag']
        new = self.__class__(self.scalar * scale_factor, new_resolution)
        return new


class ResolvedWindow(Resolved, ub.NiceRepr):
    """
    Parse a window size at a particular resolution

    Example:
        >>> from watch.utils.util_resolution import *  # NOQA
        >>> resolved_window = "128x128@10GSD"
        >>> resolved_window = ResolvedWindow.parse(resolved_window)
        >>> print('resolved_window = {}'.format(ub.urepr(resolved_window, nl=1, sv=1)))
        resolved_window = <ResolvedWindow((128, 128) @ {'mag': 10, 'unit': 'GSD'})>
        >>> resolved_window = "128  ,  128  @  10  GSD"
        >>> resolved_window = ResolvedWindow.parse(resolved_window)
        >>> print('resolved_window = {}'.format(ub.urepr(resolved_window, nl=1, sv=1)))
        resolved_window = <ResolvedWindow((128, 128) @ {'mag': 10, 'unit': 'GSD'})>
        >>> from watch.utils.util_resolution import *  # NOQA
        >>> resolved_window = "128@10GSD"
        >>> resolved_window = ResolvedWindow.parse(resolved_window)
        >>> print('resolved_window = {}'.format(ub.urepr(resolved_window, nl=1, sv=1)))
        resolved_window = <ResolvedWindow((128, 128) @ {'mag': 10, 'unit': 'GSD'})>

    """
    __transformer__ = ResolvedWindowTransformer

    def __init__(self, window, resolution):
        self.window = window
        self.resolution = resolution

    def at_resolution(self, new_resolution):
        '''
        Update the resolution

        Example:
            >>> from watch.utils.util_resolution import *  # NOQA
            >>> new_resolution = {'mag': 1, 'unit': 'GSD'}
            >>> self = ResolvedWindow.parse("128x64@10GSD")
            >>> print(self.at_resolution(new_resolution))
            >>> print(self.at_resolution({'mag': 20, 'unit': 'GSD'}))
            <ResolvedWindow((1280.0, 640.0) @ {'mag': 1, 'unit': 'GSD'})>
            <ResolvedWindow((64.0, 32.0) @ {'mag': 20, 'unit': 'GSD'})>
        '''
        scale_factor = self.resolution['mag'] / new_resolution['mag']
        w, h = self.window
        new_window = (w * scale_factor, h * scale_factor)
        new = self.__class__(new_window, new_resolution)
        return new

    def __nice__(self):
        return (f'{self.window} @ {self.resolution}')
